#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2021 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

import ast
import sys
import re

try:
    from metapensiero.pj.api import translates
except ImportError:
    pass  # metapensiero not installed

import astunparse


class psychoJSTransformer(ast.NodeTransformer):
    """PsychoJS-specific AST transformer
    """

    def visit_Name(self, node):
        # status = STOPPED => status = PsychoJS.Status.STOPPED
        if node.id in ['STARTED', 'FINISHED', 'STOPPED'] and isinstance(node.ctx, ast.Load):
            return ast.copy_location(
                ast.Attribute(
                    value=ast.Attribute(
                        value=ast.Name(id='PsychoJS', ctx=ast.Load()),
                        attr='Status',
                        ctx=ast.Load()
                    ),
                    attr=node.id,
                    ctx=ast.Load()
                ),
                node)

        # return the node by default:
        return node


class pythonTransformer(ast.NodeTransformer):
    """Python-specific AST transformer
    """

    # builtin python operations that require substitution by specific JavaScript code:
    subtitutableOperations = []

    # operations from the math python module or builtin operations that exist in JavaScript Math:
    directMathOperations = ['abs', 'min', 'max', 'ceil', 'fabs', 'floor', 'trunc', 'exp', 'log', 'log2', 'pow',
                            'sqrt', 'acos', 'asin', 'atan2', 'cos', 'sin', 'tan', 'acosh', 'asinh', 'atanh', 'cosh',
                            'sinh', 'tanh']

    def visit_BinOp(self, node):

        # formatted strings with %:
        if isinstance(node.op, ast.Mod) and isinstance(node.left, ast.Str):
            raise Exception('string formatting using % is not currently supported, please use f-strings instead')

        return node

    def visit_FormattedValue(self, node):

        # unformatted f-strings:
        if not node.format_spec:
            return node

        # formatted f-strings:
        if isinstance(node.format_spec, ast.JoinedStr) and len(node.format_spec.values) > 0 and isinstance(
                node.format_spec.values[0], ast.Str):

            # split the format:
            format = node.format_spec.values[0].s
            match = re.search(r"([0-9]*).([0-9]+)(f|i)", format)
            if not match:
                raise Exception(format + ' format is not currently supported')

            matchGroups = match.groups()
            width = matchGroups[0]
            width = int(width) if width != '' else 1
            precision = matchGroups[1]
            precision = int(precision) if precision != '' else 1
            conversion = matchGroups[2]
            value = node.value

            # prepare the conversion:
            conversionFunc = ast.Call(
                func=ast.Attribute(
                    value=ast.Name(id='Number', ctx=ast.Load()),
                    attr='parseFloat' if conversion == 'f' else 'parseInt',
                    ctx=ast.Load()
                ),
                args=[value],
                keywords=[]
            )

            # deal with precision:
            precisionCall = ast.Call(
                func=ast.Attribute(
                    value=conversionFunc,
                    attr='toPrecision',
                    ctx=ast.Load()
                ),
                args=[ast.Num(n=precision)],
                keywords=[]
            )

            # deal with width:
            widthCall = ast.Call(
                func=ast.Name(id='pad', ctx=ast.Load()),
                args=[precisionCall, ast.Num(n=width)],
                keywords=[]
            )

            # return the node:
            node.value = widthCall
            node.conversion = -1
            node.format_spec = None

            return node

        raise Exception('formatted f-string are not all supported at the moment')

    def visit_Call(self, node):

        # if the call node has arguments, transform them first:
        nbArgs = len(node.args)
        for i in range(0, nbArgs):
            node.args[i] = pythonTransformer().visit(node.args[i])

        # operations with module prefix, e.g. a = math.fabs(-1.2)
        if isinstance(node.func, ast.Attribute) and isinstance(node.func.value, ast.Name):
            prefix = node.func.value.id
            attribute = node.func.attr

            if prefix == 'math':
                return self.mathTransform(attribute, node.args)
            else:
                return node

        # operations without prefix:
        elif isinstance(node.func, ast.Name):
            attribute = node.func.id

            # check whether this is a math operation:
            mathNode = self.mathTransform(attribute, node.args)
            if mathNode:
                return mathNode

        # string.format(args):
        elif isinstance(node.func, ast.Attribute) and isinstance(node.func.value,
                                                                 ast.Str) and node.func.attr == 'format':
            raise Exception('format() is not supported at the moment, please use f-strings instead')

        # return the node by default:
        return node

    def mathTransform(self, attribute, args):

        # operations from the math python module or builtin operations that exist in JavaScript Math:
        if attribute in self.directMathOperations:
            func = ast.Attribute(
                value=ast.Name(id='Math', ctx=ast.Load()),
                attr=attribute,
                ctx=ast.Load()
            )

            return ast.Call(
                func=func,
                args=args,
                keywords=[]
            )

        # substitutable operations, e.g. a = sum(b,c) => a = [b,c].reduce( function(x,y) { return x+y: })
        elif attribute in self.subtitutableOperations:
            # a = sum(b,c) => a = [b,c].reduce( function(x,y) { return x+y: })
            pass  # removed for now in preference for creating the func in utils

        else:
            return None


class pythonAddonVisitor(ast.NodeVisitor):
    # operations that require an addon:
    addonOperations = ['list', 'pad']

    def __init__(self):
        self.addons = []

    def visit_Call(self, node):
        if isinstance(node.func, ast.Name) and node.func.id in self.addonOperations:
            self.addons.append(node.func.id)


def transformNode(astNode):
    """Transform the input AST

    Args:
        astNode (ast.Node): the input AST

    Returns:
        ast.Node: transformed AST
    """

    # deal with PsychoJS specific changes:
    psychoJSTransformedNode = psychoJSTransformer().visit(astNode)

    # deal with python specific changes:
    pythonBuiltinTransformedNode = pythonTransformer().visit(psychoJSTransformedNode)

    # look for operations requiring an addon:
    visitor = pythonAddonVisitor()
    visitor.visit(psychoJSTransformedNode)

    return pythonBuiltinTransformedNode, visitor.addons


def transformPsychoJsCode(psychoJsCode, addons):
    """Transform the input PsychoJS code.

    Args:
        psychoJsCode (str): the input PsychoJS JavaScript code

    Returns:
        (str) the transformed code
    """

    transformedPsychoJSCode = ''

    # add addons on a need-for basis:
    if 'list' in addons:
        transformedPsychoJSCode += """
        // add-on: list(s: string): string[]
        function list(s) {
            // if s is a string, we return a list of its characters
            if (typeof s === 'string')
                return s.split('');
            else
                // otherwise we return s:
                return s;
        }

        """

    if 'pad' in addons:
        transformedPsychoJSCode += r"""
        // add-on: pad(n: number, width: number): string
        function pad(n, width) {
            width = width || 2;
            integerPart = Number.parseInt(n);
            decimalPart = (n+'').match(/\.[0-9]*/);
            if (!decimalPart)
                decimalPart = '';
            return (integerPart+'').padStart(width,'0') + decimalPart;
        }

        """

    lines = psychoJsCode.splitlines()

    # remove the initial variable declarations, unless it is for _pj:
    if lines[0].find('var _pj;') == 0:
        transformedPsychoJSCode = 'var _pj;\n'
        startIndex = 1
    else:
        startIndex = 0

    if lines[startIndex].find('var') == 0:
        startIndex += 1

    for index in range(startIndex, len(lines)):
        transformedPsychoJSCode += lines[index]
        transformedPsychoJSCode += '\n'

    return transformedPsychoJSCode


def translatePythonToJavaScript(psychoPyCode):
    """Translate PsychoPy python code into PsychoJS JavaScript code.

    Args:
        psychoPyCode (str): the input PsychoPy python code

    Returns:
        str: the PsychoJS JavaScript code

    Raises:
        (Exception): whenever a step of the translation process failed
    """

    # get the Abstract Syntax Tree (AST)
    # this checks that the code is valid python
    try:
        astNode = ast.parse(psychoPyCode)
    # print('>>> AST node: ' + ast.dump(astNode))
    except Exception as error:
        raise Exception('unable to parse the PsychoPy code into an abstract syntax tree: ' + str(error))

    # transform the AST by making PsychoJS-specific substitutions and dealing with python built-ins:
    try:
        transformedAstNode, addons = transformNode(astNode)
    # print('>>> transformed AST node: ' + ast.dump(transformedAstNode))
    # print('>>> addons: ' + str(addons))
    except Exception as error:
        raise Exception('unable to transform the abstract syntax tree: ' + str(error))

    # turn the transformed AST into code:
    try:
        transformedPsychoPyCode = astunparse.unparse(transformedAstNode)
    # print('>>> transformed PsychoPy code:\n' + transformedPsychoPyCode)
    except Exception as error:
        raise Exception('unable to turn the transformed abstract syntax tree back into code: ' + str(error))

    # translate the python code into JavaScript code:
    try:
        psychoJsCode, psychoJsSourceMap = translates(transformedPsychoPyCode, enable_es6=True)
    # print('>>> PsychoJS code:\n' + psychoJsCode)
    except Exception as error:
        raise Exception(
            'unable to translate the transformed PsychoPy code into PsychoJS JavaScript code: ' + str(error))

    # transform the JavaScript code:
    try:
        transformedPsychoJsCode = transformPsychoJsCode(psychoJsCode, addons)
    except Exception as error:
        raise Exception('unable to transform the PsychoJS JavaScript code: ' + str(error))

    return transformedPsychoJsCode


def main(argv=None):
    """Read PsychoPy code from the command line and translate it into PsychoJS code.
    """

    # other read PsychoPy code from the command line:
    print('Enter PsychoPY code (finish with Ctrl+Z):')
    psychoPyCode = sys.stdin.read()

    # translate it to PsychoJS:
    psychoJSCode = translatePythonToJavaScript(psychoPyCode)
    print('>>> translated PsychoJS code:\n' + psychoJSCode)


if __name__ == "__main__":
    main()



from __future__ import print_function
from past.builtins import unicode

from builtins import object
import os
import io
import pytest
import warnings

from psychopy import constants
from psychopy.experiment import getAllComponents
from psychopy import experiment
from pkg_resources import parse_version

# use "python genComponsTemplate.py --out" to generate a new profile to test against
#   = analogous to a baseline image to compare screenshots
# motivation: catch deviations introduced during refactoring

# what reference to use?
profile = 'componsTemplate.txt'

# always ignore hints, labels, and categories. other options:
# should it be ok or an error if the param[field] order differs from the profile?
ignoreOrder = True

# ignore attributes that are there because inherit from object
ignoreObjectAttribs = True
ignoreList = ['<built-in method __', "<method-wrapper '__", '__slotnames__:']


# profile is not platform specific, which can trigger false positives.
# allowedVals can differ across platforms or with prefs:
ignoreParallelOutAddresses = True

@pytest.mark.components
class TestComponents(object):
    @classmethod
    def setup_class(cls):
        cls.exp = experiment.Experiment() # create once, not every test
        cls.here = os.path.abspath(os.path.dirname(__file__))
        cls.baselineProfile = os.path.join(cls.here, profile)

        # should not need a wx.App with fetchIcons=False
        try:
            cls.allComp = getAllComponents(fetchIcons=False)
        except Exception:
            import wx
            if parse_version(wx.__version__) < parse_version('2.9'):
                tmpApp = wx.PySimpleApp()
            else:
                tmpApp = wx.App(False)
            try:
                from psychopy.app import localization
            except Exception:
                pass  # not needed if can't import it
            cls.allComp = getAllComponents(fetchIcons=False)

    def setup(self):
        """This setup is done for each test individually
        """
        pass

    def teardown(self):
        pass

    def test_component_attribs(self):
        with io.open(self.baselineProfile, 'r', encoding='utf-8-sig') as f:
            target = f.read()
        targetLines = target.splitlines()
        targetTag = {}
        for line in targetLines:
            try:
                t, val = line.split(':',1)
                targetTag[t] = val
            except ValueError:
                # need more than one value to unpack; this is a weak way to
                # handle multi-line default values, eg TextComponent.text.default
                targetTag[t] += '\n' + line  # previous t value

        param = experiment.Param('', '')  # want its namespace
        ignore = ['__doc__', '__init__', '__module__', '__str__', 'next',
                  '__unicode__', '__native__', '__nonzero__', '__long__']

        # these are for display only (cosmetic) and can end up being localized
        # so typically do not want to check during automated testing, at least
        # not when things are still new-ish and subject to change:
        ignore += ['hint',
                   'label',  # comment-out to compare labels when checking
                   'categ',
                   'next',
                   'dollarSyntax',
                   ]
        for field in dir(param):
            if field.startswith("__"):
                ignore.append(field)
        fields = set(dir(param)).difference(ignore)

        mismatched = []
        for compName in sorted(self.allComp):
            comp = self.allComp[compName](parentName='x', exp=self.exp)
            order = '%s.order:%s' % (compName, eval("comp.order"))

            if order+'\n' not in target:
                tag = order.split(':',1)[0]
                try:
                    mismatch = order + ' <== ' + targetTag[tag]
                except (IndexError, KeyError): # missing
                    mismatch = order + ' <==> NEW (no matching param in the reference profile)'
                print(mismatch.encode('utf8'))

                if not ignoreOrder:
                    mismatched.append(mismatch)

            for parName in comp.params:
                # default is what you get from param.__str__, which returns its value
                if not constants.PY3:
                    if isinstance(comp.params[parName].val, unicode):
                        comp.params[parName].val = comp.params[parName].val.encode('utf8')
                default = '%s.%s.default:%s' % (compName, parName, comp.params[parName])
                lineFields = []
                for field in fields:
                    if parName == 'name' and field == 'updates':
                        continue
                        # ignore b/c never want to change the name *during a running experiment*
                        # the default name.updates varies across components: need to ignore or standardize
                    f = '%s.%s.%s:%s' % (compName, parName, field, eval("comp.params[parName].%s" % field))
                    lineFields.append(f)

                for line in [default] + lineFields:
                    # some attributes vary by machine so don't check those
                    if line.startswith('ParallelOutComponent.address') and ignoreParallelOutAddresses:
                        continue
                    elif line.startswith('SettingsComponent.OSF Project ID.allowedVals'):
                        continue
                    elif ('SettingsComponent.Use version.allowedVals' in line or
                        'SettingsComponent.Use version.__dict__' in line):
                        # versions available on travis-ci are only local
                        continue
                    origMatch = line+'\n' in target
                    lineAlt = (line.replace(":\'", ":u'")
                                    .replace("\\\\","\\")
                                    .replace("\\'", "'"))
                    # start checking params
                    if not (line+'\n' in target
                            or lineAlt+'\n' in target):
                        # mismatch, so report on the tag from orig file
                        # match checks tag + multi-line, because line is multi-line and target is whole file
                        tag = line.split(':',1)[0]
                        try:
                            mismatch = line + ' <== ' + targetTag[tag]
                        except KeyError: # missing
                            mismatch = line + ' <==> NEW (no matching param in the reference profile)'

                        # ignore attributes that inherit from object:

                        if ignoreObjectAttribs:
                            for item in ignoreList:
                                if item in mismatch:
                                    break
                            else:
                                mismatched.append(mismatch)
                        else:
                            mismatched.append(mismatch)

        for mismatch in mismatched:
            warnings.warn("Non-identical Builder Param: {}".format(mismatch))


@pytest.mark.components
def test_flip_before_shutdown_in_settings_component():
    exp = experiment.Experiment()
    script = exp.writeScript()

    assert 'Flip one final time' in script



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Demo of some ways to use class psychopy.info.RunTimeInfo()
to obtain current system and other data at run-time.
"""

from __future__ import absolute_import, division, print_function

from builtins import str
from psychopy import visual, logging, core
import psychopy.info

# author and version are used in the demo, in the way you might in your experiment.
# They are expected to be at the top of the script that calls RunTimeInfo()),
# with a string literal assigned to them (no variables).
# double-quotes will be silently removed, single quotes will be left, eg, O'Connor
__author__ = """Jeremy "R." Gray"""
__version__ = "v1.0.a"

# When creating an experiment, first define your window (& monitor):
win = visual.Window(fullscr=False, size=[200, 200], monitor='testMonitor')
win.recordFrameIntervals = True
logging.console.setLevel(logging.DEBUG)

# Then gather run-time info. All parameters are optional:
runInfo = psychopy.info.RunTimeInfo(
        # if you specify author and version here, it overrides the automatic detection of __author__ and __version__ in your script
        # author=' < your name goes here, plus whatever you like, e.g., your lab or contact info > ',
        # version=" < your experiment version info > ",
        win=win,    #  # a psychopy.visual.Window() instance; None = default temp window used; False = no win, no win.flips()
        refreshTest='grating',  #  # None, True, or 'grating' (eye-candy to avoid a blank screen)
        verbose=True,  #  # True means report on everything
        userProcsDetailed=True,  #  # if verbose and userProcsDetailed, return (command, process-ID) of the user's processes
        )
win.close()

print("""
System and other run-time details are now saved in "runInfo", a dict-like object. You have to decide
what to do with it.

"print(runInfo)" will give you the same as "print(str(runInfo))". This format is intended to be useful
for writing to a data file in a human readable form:""")
print(runInfo)

print("If that's more detail than you want, try: runInfo = info.RunTimeInfo(..., verbose=False, ...).")

# To get the same info in python syntax, use "print(repr(info))".
# You could write this format into a data file, and it's fairly readable.
# And because its python syntax you could later simply
# import your data file into python to reconstruct the dict.

print("\nYou can extract single items from info, using keys, e.g.:")
print("  psychopyVersion = %s" % runInfo['psychopyVersion'])
if "windowRefreshTimeAvg_ms" in runInfo:
    print("or from the test of the screen refresh rate:")
    print("  %.2f ms = average refresh time" % runInfo["windowRefreshTimeAvg_ms"])
    print("  %.3f ms = standard deviation" % runInfo["windowRefreshTimeSD_ms"])

    # Once you have run-time info, you can fine-tune things with the values, prior to running your experiment.
    refreshSDwarningLevel_ms = 0.20  # ms
    if runInfo["windowRefreshTimeSD_ms"] > refreshSDwarningLevel_ms:
        print("\nThe variability of the refresh rate is sort of high (SD > %.2f ms)." % (refreshSDwarningLevel_ms))
        # and here you could prompt the user with suggestions, possibly based on other info:
        if runInfo["windowIsFullScr"]:
            print("Your window is full-screen, which is good for timing.")
            print('Possible issues: internet / wireless? bluetooth? recent startup (not finished)?')
            if len(runInfo['systemUserProcFlagged']):
                print('other programs running? (command, process-ID):' + str(runInfo['systemUserProcFlagged']))
        else:
            print("""Try defining the window as full-screen (it's not currently), i.e. at the top of the demo change to:
    win = visual.Window((800, 600), fullscr=True, ...
and re-run the demo.""")

print("""
(NB: The visual is not the demo! Scroll up to see the text output.)""")

win.close()
core.quit()

# The contents of this file are in the public domain.



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2021 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

"""Provides functions for logging error and other messages to one or more
files and/or the console, using python's own logging module. Some warning
messages and error messages are generated by PsychoPy itself. The user can
generate more using the functions in this module.

There are various levels for logged messages with the following order of
importance: ERROR, WARNING, DATA, EXP, INFO and DEBUG.

When setting the level for a particular log target (e.g. LogFile)
the  user can set the minimum level that is required
for messages to enter the log. For example, setting a level of INFO will
result in INFO, EXP, DATA, WARNING and ERROR messages to be recorded but not
DEBUG messages.

By default, PsychoPy will record messages of WARNING level and above to
the console. The user can silence that by setting it to receive only CRITICAL
messages, (which PsychoPy doesn't use) using the commands::

    from psychopy import logging
    logging.console.setLevel(logging.CRITICAL)

"""

# Much of the code below is based conceptually, if not syntactically, on the
# python logging module but it's simpler (no threading) and maintaining a
# stack of log entries for later writing (don't want files written while
# drawing)

from __future__ import absolute_import, print_function

from builtins import object
from past.builtins import basestring
from os import path
import atexit
import sys
import codecs
import locale
from psychopy import clock
from psychopy.constants import PY3

_packagePath = path.split(__file__)[0]

CRITICAL = 50
FATAL = CRITICAL
ERROR = 40
DEPRECATION = ERROR
WARNING = 30
WARN = WARNING
DATA = 25  # will be a custom level
EXP = 22  # info about the experiment might be less important than data info?
INFO = 20
DEBUG = 10
NOTSET = 0

_levelNames = {
    CRITICAL: 'CRITICAL',
    ERROR: 'ERROR',
    WARNING: 'WARNING',
    DATA: 'DATA',
    EXP: 'EXP',
    INFO: 'INFO',
    DEBUG: 'DEBUG',
    NOTSET: 'NOTSET',
    'CRITICAL': CRITICAL,
    'ERROR': ERROR,
    'DEPRECATION': DEPRECATION,
    'DATA': DATA,
    'EXP': EXP,
    'WARN': WARNING,
    'WARNING': WARNING,
    'INFO': INFO,
    'DEBUG': DEBUG,
    'NOTSET': NOTSET}

_prefEncoding = locale.getpreferredencoding()

def getLevel(level):
    """Return the textual representation of logging level 'level'.

    If the level is one of the predefined levels (CRITICAL, ERROR, WARNING,
    INFO, DEBUG) then you get the corresponding string. If you have
    associated levels with names using addLevelName then the name you have
    associated with 'level' is returned.

    If a numeric value corresponding to one of the defined levels is passed
    in, the corresponding string representation is returned.

    Otherwise, the string "Level %s" % level is returned.
    """
    return _levelNames.get(level, "Level %s" % level)


def addLevel(level, levelName):
    """Associate 'levelName' with 'level'.

    This is used when converting levels to text during message formatting.
    """
    _levelNames[level] = levelName
    _levelNames[levelName] = level


# global defaultClock
defaultClock = clock.monotonicClock


def setDefaultClock(clock):
    """Set the default clock to be used to reference all logging times.
    Must be a :class:`psychopy.core.Clock` object. Beware that if you
    reset the clock during the experiment then the resets will be
    reflected here. That might be useful if you want your logs to be
    reset on each trial, but probably not.
    """
    global defaultClock
    defaultClock = clock


class _LogEntry(object):

    def __init__(self, level, message, t=None, obj=None):
        super(_LogEntry, self).__init__()
        try:
            "%0.4f" % (t)
        except (ValueError, TypeError):
            raise ValueError("Value \"%s\" of log message \"%s\" could not be coerced to string from numeric" % (t, message))
        self.t = t
        self.t_ms = t * 1000
        self.level = level
        self.levelname = getLevel(level)
        self.message = message
        self.obj = obj


class LogFile(object):
    """A text stream to receive inputs from the logging system
    """

    def __init__(self, f=None, level=WARNING, filemode='a', logger=None,
                 encoding='utf8'):
        """Create a log file as a target for logged entries of a given level

        :parameters:

            - f:
                this could be a string to a path, that will be created if it
                doesn't exist. Alternatively this could be a file object,
                sys.stdout or any object that supports .write() and .flush()
                methods

            - level:
                The minimum level of importance that a message must have
                to be logged by this target.

            - filemode: 'a', 'w'
                Append or overwrite existing log file

        """
        super(LogFile, self).__init__()
        # work out if this is a filename or a stream to write to
        if f is None:
            self.stream = 'stdout'
        elif hasattr(f, 'write'):
            self.stream = f
        elif isinstance(f, basestring):
            self.stream = codecs.open(f, filemode, encoding)
        self.level = level
        if logger is None:
            logger = root
        # Can not use weak ref to logger, as sometimes this class
        # instance would be gc'ed before _Logger.__del__
        # was complete (running .flush()).
        # This was causing following error when script closed:
        #     Exception AttributeError: "'NoneType' object has no
        #       attribute 'stdout'" in
        #     <bound method _Logger.__del__ of
        #       <psychopy.logging._Logger instance at 0x102e0d878>> ignored
        # So instead free logger ref in __del__ of this class,
        # so we know any log backlog can be flushed before it is gc'ed.
        self.logger = logger

        self.logger.addTarget(self)

    def setLevel(self, level):
        """Set a new minimal level for the log file/stream
        """
        if type(level) is not int:
            raise TypeError("LogFile.setLevel() should be given an int, which"
                            "is usually one of logging.INFO (not logging.info)")
        self.level = level
        self.logger._calcLowestTarget()

    def write(self, txt):
        """Write directly to the log file (without using logging functions).
        Useful to send messages that only this file receives
        """
        # find the current stdout if we're the console logger
        if self.stream == 'stdout':
            if PY3:
                stream = sys.stdout
            else:
                stream = codecs.getwriter(_prefEncoding)(sys.stdout)
        else:
            stream = self.stream
        # try to write
        try:
            stream.write(txt)
        except UnicodeEncodeError as e:  # incompatible encoding of stdout?
            try:
                if hasattr(stream, 'reconfigure'):
                    stream.reconfigure(encoding='utf-8')
                elif stream == sys.stdout:
                    # try opening sys.stdout manually as a file
                    sys.stdout = stream = open(sys.stdout.fileno(), mode='w',
                                               encoding='utf-8', buffering=1)
                stream.write(txt)  # try again with the new encoding
            except Exception:
                print('Failed to reconfigure logger output encoding', e)

        try:
            stream.flush()
        except Exception:
            pass


class _Logger(object):
    """Maintains a set of log targets (text streams such as files of stdout)

    self.targets is a list of dicts {'stream':stream, 'level':level}

    """

    def __init__(self, format="%(t).4f \t%(levelname)s \t%(message)s"):
        """The string-formatted elements %(xxxx)f can be used, where
        each xxxx is an attribute of the LogEntry.
        e.g. t, t_ms, level, levelname, message
        """
        super(_Logger, self).__init__()
        self.targets = []
        self.flushed = []
        self.toFlush = []
        self.format = format
        self.lowestTarget = 50

    def __del__(self):
        self.flush()
        # unicode logged to coder output window can cause logger failure, with
        # error message pointing here. this is despite it being ok to log to
        # terminal or Builder output. proper fix: fix coder unicode bug #97
        # (currently closed)

    def addTarget(self, target):
        """Add a target, typically a :class:`~log.LogFile` to the logger
        """
        self.targets.append(target)
        self._calcLowestTarget()

    def removeTarget(self, target):
        """Remove a target, typically a :class:`~log.LogFile` from the logger
        """
        if target in self.targets:
            self.targets.remove(target)
        self._calcLowestTarget()

    def _calcLowestTarget(self):
        self.lowestTarget = 50
        for target in self.targets:
            self.lowestTarget = min(self.lowestTarget, target.level)

    def log(self, message, level, t=None, obj=None):
        """Add the `message` to the log stack at the appropriate `level`

        If no relevant targets (files or console) exist then the message is
        simply discarded.
        """
        # check for at least one relevant logger
        if level < self.lowestTarget:
            return
        # check time
        if t is None:
            global defaultClock
            t = defaultClock.getTime()
        # add message to list
        self.toFlush.append(
            _LogEntry(t=t, level=level, message=message, obj=obj))

    def flush(self):
        """Process all current messages to each target
        """
        # loop through targets then entries in toFlush
        # so that stream.flush can be called just once
        formatted = {}  # keep a dict - so only do the formatting once
        for target in self.targets:
            for thisEntry in self.toFlush:
                if thisEntry.level >= target.level:
                    if not thisEntry in formatted:
                        # convert the entry into a formatted string
                        formatted[thisEntry] = self.format % thisEntry.__dict__
                    target.write(formatted[thisEntry] + '\n')
            if hasattr(target.stream, 'flush'):
                target.stream.flush()
        # finished processing entries - move them to self.flushed
        self.flushed.extend(self.toFlush)
        self.toFlush = []  # a new empty list

root = _Logger()
console = LogFile()


def flush(logger=root):
    """Send current messages in the log to all targets
    """
    logger.flush()

# make sure this function gets called as python closes
atexit.register(flush)


def critical(msg, t=None, obj=None):
    """log.critical(message)
    Send the message to any receiver of logging info (e.g. a LogFile)
    of level `log.CRITICAL` or higher
    """
    root.log(msg, level=CRITICAL, t=t, obj=obj)
fatal = critical


def error(msg, t=None, obj=None):
    """log.error(message)

    Send the message to any receiver of logging info (e.g. a LogFile)
    of level `log.ERROR` or higher
    """
    root.log(msg, level=ERROR, t=t, obj=obj)


def warning(msg, t=None, obj=None):
    """log.warning(message)

    Sends the message to any receiver of logging info (e.g. a LogFile)
    of level `log.WARNING` or higher
    """
    root.log(msg, level=WARNING, t=t, obj=obj)
warn = warning


def data(msg, t=None, obj=None):
    """Log a message about data collection (e.g. a key press)

    usage::
        log.data(message)

    Sends the message to any receiver of logging info (e.g. a LogFile)
    of level `log.DATA` or higher
    """
    root.log(msg, level=DATA, t=t, obj=obj)


def exp(msg, t=None, obj=None):
    """Log a message about the experiment
    (e.g. a new trial, or end of a stimulus)

    usage::
        log.exp(message)

    Sends the message to any receiver of logging info (e.g. a LogFile)
    of level `log.EXP` or higher
    """
    root.log(msg, level=EXP, t=t, obj=obj)


def info(msg, t=None, obj=None):
    """Log some information - maybe useful, maybe not

    usage::
        log.info(message)

    Sends the message to any receiver of logging info (e.g. a LogFile)
    of level `log.INFO` or higher
    """
    root.log(msg, level=INFO, t=t, obj=obj)


def debug(msg, t=None, obj=None):
    """Log a debugging message (not likely to be wanted once
    experiment is finalised)

    usage::
        log.debug(message)

    Sends the message to any receiver of logging info (e.g. a LogFile)
    of level `log.DEBUG` or higher
    """
    root.log(msg, level=DEBUG, t=t, obj=obj)


def log(msg, level, t=None, obj=None):
    """Log a message

    usage::
        log(level, msg, t=t, obj=obj)

    Log the msg, at a  given level on the root logger
    """
    root.log(msg, level=level, t=t, obj=obj)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2021 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

"""To build simple dialogues etc. (requires wxPython)
"""

from __future__ import absolute_import, print_function

from builtins import str
from builtins import super
from builtins import range
from psychopy import logging
import wx
import numpy
import os
from psychopy.localization import _translate
from pkg_resources import parse_version

OK = wx.ID_OK

thisVer = parse_version(wx.__version__)

def ensureWxApp():
    # make sure there's a wxApp prior to showing a gui, e.g., for expInfo
    # dialog
    try:
        wx.Dialog(None, -1)  # not shown; FileDialog gives same exception
        return True
    except wx._core.PyNoAppError:
        if thisVer < parse_version('2.9'):
            return wx.PySimpleApp()
        elif thisVer >= parse_version('4.0') and thisVer < parse_version('4.1'):
            raise Exception(
                    "wx>=4.0 clashes with pyglet and making it unsafe "
                    "as a PsychoPy gui helper. Please install PyQt (4 or 5)"
                    " or wxPython3 instead.")
        else:
            return wx.App(False)


class Dlg(wx.Dialog):
    """A simple dialogue box. You can add text or input boxes
    (sequentially) and then retrieve the values.

    see also the function *dlgFromDict* for an **even simpler** version

    **Example:**    ::

        from psychopy import gui

        myDlg = gui.Dlg(title="JWP's experiment")
        myDlg.addText('Subject info')
        myDlg.addField('Name:')
        myDlg.addField('Age:', 21)
        myDlg.addText('Experiment Info')
        myDlg.addField('Grating Ori:',45)
        myDlg.addField('Group:', choices=["Test", "Control"])
        myDlg.show()  # show dialog and wait for OK or Cancel
        if myDlg.OK:  # then the user pressed OK
            thisInfo = myDlg.data
            print(thisInfo)
        else:
            print('user cancelled')
    """

    def __init__(self, title=_translate('PsychoPy dialogue'),
                 pos=wx.DefaultPosition, size=wx.DefaultSize,
                 style=wx.DEFAULT_DIALOG_STYLE | wx.DIALOG_NO_PARENT,
                 labelButtonOK=_translate(" OK "),
                 labelButtonCancel=_translate(" Cancel ")):
        style = style | wx.RESIZE_BORDER
        global app  # avoid recreating for every gui
        if pos is None:
            pos = wx.DefaultPosition
        app = ensureWxApp()
        super().__init__(parent=None, id=-1, title=title, style=style, pos=pos)
        self.inputFields = []
        self.inputFieldTypes = []
        self.inputFieldNames = []
        self.data = []
        # prepare a frame in which to hold objects
        self.sizer = wx.BoxSizer(wx.VERTICAL)
        # self.addText('')  # insert some space at top of dialogue
        self.pos = pos
        self.labelButtonOK = labelButtonOK
        self.labelButtonCancel = labelButtonCancel

    def addText(self, text, color=''):
        # the horizontal extent can depend on the locale and font in use:
        font = self.GetFont()
        dc = wx.WindowDC(self)
        dc.SetFont(font)
        textWidth, textHeight = dc.GetTextExtent(text)
        textLength = wx.Size(textWidth + 50, textHeight)

        _style = wx.ALIGN_CENTER_VERTICAL | wx.ALIGN_CENTER_HORIZONTAL
        myTxt = wx.StaticText(self, -1, label=text, style=_style,
                              size=textLength)
        if len(color):
            myTxt.SetForegroundColour(color)
        self.sizer.Add(myTxt, 1, wx.ALIGN_CENTER)

    def addField(self, label='', initial='', color='', choices=None, tip=''):
        """Adds a (labelled) input field to the dialogue box, optional text
        color and tooltip. Returns a handle to the field (but not to the
        label). If choices is a list or tuple, it will create a dropdown
        selector.
        """
        self.inputFieldNames.append(label)
        if choices:
            self.inputFieldTypes.append(str)
        else:
            self.inputFieldTypes.append(type(initial))
        if type(initial) == numpy.ndarray:
            initial = initial.tolist()  # convert numpy arrays to lists
        container = wx.GridSizer(cols=2, vgap=0, hgap=10)
        # create label
        font = self.GetFont()
        dc = wx.WindowDC(self)
        dc.SetFont(font)
        labelWidth, labelHeight = dc.GetTextExtent(label)
        labelLength = wx.Size(labelWidth + 16, labelHeight)
        inputLabel = wx.StaticText(self, -1, label,
                                   size=labelLength,
                                   style=wx.ALIGN_RIGHT)
        if len(color):
            inputLabel.SetForegroundColour(color)
        _style = wx.ALIGN_CENTER_VERTICAL | wx.ALIGN_RIGHT
        container.Add(inputLabel, 1, _style)
        # create input control
        if type(initial) == bool:
            inputBox = wx.CheckBox(self, -1)
            inputBox.SetValue(initial)
        elif not choices:
            inputWidth, inputHeight = dc.GetTextExtent(str(initial))
            inputLength = wx.Size(max(50, inputWidth + 16),
                                  max(25, inputHeight + 8))
            inputBox = wx.TextCtrl(self, -1, str(initial),
                                   size=inputLength)
        else:
            inputBox = wx.Choice(self, -1,
                                 choices=[str(option)
                                          for option in list(choices)])
            # Somewhat dirty hack that allows us to treat the choice just like
            # an input box when retrieving the data
            inputBox.GetValue = inputBox.GetStringSelection
            initial = choices.index(initial) if initial in choices else 0
            inputBox.SetSelection(initial)
        if len(color):
            inputBox.SetForegroundColour(color)
        if len(tip):
            inputBox.SetToolTip(wx.ToolTip(tip))

        container.Add(inputBox, 1, wx.ALIGN_CENTER_VERTICAL)
        self.sizer.Add(container, 1, wx.ALIGN_CENTER)

        self.inputFields.append(inputBox)  # store this to get data back on OK
        return inputBox

    def addFixedField(self, label='', value='', tip=''):
        """Adds a field to the dialogue box (like addField) but the
        field cannot be edited. e.g. Display experiment version.
        tool-tips are disabled (by wx).
        """
        thisField = self.addField(label, value, color='Gray', tip=tip)
        # wx disables tooltips too; we pass them in anyway
        thisField.Disable()
        return thisField

    def display(self):
        """Presents the dialog and waits for the user to press OK or CANCEL.

        If user presses OK button, function returns a list containing the
        updated values coming from each of the input fields created.
        Otherwise, None is returned.

        :return: self.data
        """
        # add buttons for OK and Cancel
        buttons = wx.BoxSizer(wx.HORIZONTAL)
        OK = wx.Button(self, wx.ID_OK, self.labelButtonOK)
        OK.SetDefault()

        buttons.Add(OK)
        CANCEL = wx.Button(self, wx.ID_CANCEL, self.labelButtonCancel)
        buttons.Add(CANCEL)
        self.sizer.Add(buttons, 1, flag=wx.ALIGN_RIGHT, border=5)

        self.SetSizerAndFit(self.sizer)
        if self.pos is None:
            self.Center()
        if self.ShowModal() == wx.ID_OK:
            self.data = []
            # get data from input fields
            for n in range(len(self.inputFields)):
                thisName = self.inputFieldNames[n]
                thisVal = self.inputFields[n].GetValue()
                thisType = self.inputFieldTypes[n]
                # try to handle different types of input from strings
                logging.debug("%s: %s" % (self.inputFieldNames[n],
                                          str(thisVal)))
                if thisType in (tuple, list, float, int):
                    # probably a tuple or list
                    exec("self.data.append(" + thisVal + ")")  # evaluate it
                elif thisType == numpy.ndarray:
                    exec("self.data.append(numpy.array(" + thisVal + "))")
                elif thisType in (str, bool):
                    self.data.append(thisVal)
                else:
                    logging.warning('unknown type:' + self.inputFieldNames[n])
                    self.data.append(thisVal)
            self.OK = True
        else:
            self.OK = False
        self.Destroy()
        if self.OK:
            return self.data

    def show(self):
        """Presents the dialog and waits for the user to press either
        OK or CANCEL.

        When they do, dlg.OK will be set to True or False (according to
        which button they pressed. If OK==True then dlg.data will be
        populated with a list of values coming from each of the input
        fields created.
        """
        return self.display()

class DlgFromDict(Dlg):
    """Creates a dialogue box that represents a dictionary of values.
    Any values changed by the user are change (in-place) by this
    dialogue box.

    Parameters
    ----------

    sortKeys : bool
        Whether the dictionary keys should be ordered alphabetically
        for displaying.

    copyDict : bool
        If False, modify ``dictionary`` in-place. If True, a copy of
        the dictionary is created, and the altered version (after
        user interaction) can be retrieved from
        :attr:~`psychopy.gui.DlgFromDict.dictionary`.

    show : bool
        Whether to immediately display the dialog upon instantiation.
         If False, it can be displayed at a later time by calling
         its `show()` method.

    e.g.:

    ::

        info = {'Observer':'jwp', 'GratingOri':45,
                'ExpVersion': 1.1, 'Group': ['Test', 'Control']}
        infoDlg = gui.DlgFromDict(dictionary=info,
                    title='TestExperiment', fixed=['ExpVersion'])
        if infoDlg.OK:
            print(info)
        else:
            print('User Cancelled')

    In the code above, the contents of *info* will be updated to the values
    returned by the dialogue box.

    If the user cancels (rather than pressing OK),
    then the dictionary remains unchanged. If you want to check whether
    the user hit OK, then check whether DlgFromDict.OK equals
    True or False

    See GUI.py for a usage demo, including order and tip (tooltip).
    """

    def __init__(self, dictionary, title='', fixed=None, order=None, tip=None,
                 sortKeys=True, copyDict=False, show=True,
                 sort_keys=None, copy_dict=None):
        # We don't explicitly check for None identity
        # for backward-compatibility reasons.
        if not fixed:
            fixed = []
        if not order:
            order = []
        if not tip:
            tip = dict()

        # app = ensureWxApp() done by Dlg
        super().__init__(title)

        if copyDict:
            self.dictionary = dictionary.copy()
        else:
            self.dictionary = dictionary

        self._keys = list(self.dictionary.keys())

        if sortKeys:
            self._keys.sort()
        if order:
            self._keys = list(order) + list(set(self._keys).difference(set(order)))

        types = dict()

        for field in self._keys:
            types[field] = type(self.dictionary[field])
            tooltip = ''
            if field in tip:
                tooltip = tip[field]
            if field in fixed:
                self.addFixedField(field, self.dictionary[field], tip=tooltip)
            elif type(self.dictionary[field]) in [list, tuple]:
                self.addField(field, choices=self.dictionary[field],
                              tip=tooltip)
            else:
                self.addField(field, self.dictionary[field], tip=tooltip)

        if show:
            self.show()

    def show(self):
        """Display the dialog.
        """
        super().show()
        if self.OK:
            for n, thisKey in enumerate(self._keys):
                self.dictionary[thisKey] = self.data[n]


def fileSaveDlg(initFilePath="", initFileName="",
                prompt=_translate("Select file to save"),
                allowed=None):
    """A simple dialogue allowing write access to the file system.
    (Useful in case you collect an hour of data and then try to
    save to a non-existent directory!!)

    :parameters:

        initFilePath: string
            default file path on which to open the dialog

        initFileName: string
            default file name, as suggested file

        prompt: string (default "Select file to open")
            can be set to custom prompts

        allowed: string
            A string to specify file filters.
            e.g. "BMP files (*.bmp)|*.bmp|GIF files (*.gif)|*.gif"
            See http://www.wxpython.org/docs/api/wx.FileDialog-class.html
            for further details

    If initFilePath or initFileName are empty or invalid then
    current path and empty names are used to start search.

    If user cancels the None is returned.
    """
    if allowed is None:
        allowed = "All files (*.*)|*.*"
        # "txt (*.txt)|*.txt"
        # "pickled files (*.pickle, *.pkl)|*.pickle"
        # "shelved files (*.shelf)|*.shelf"
    global app  # avoid recreating for every gui
    app = ensureWxApp()
    dlg = wx.FileDialog(None, prompt, initFilePath,
                        initFileName, allowed, wx.FD_SAVE)
    if dlg.ShowModal() == OK:
        # get names of images and their directory
        outName = dlg.GetFilename()
        outPath = dlg.GetDirectory()
        dlg.Destroy()
        # tmpApp.Destroy()  # this causes an error message for some reason
        fullPath = os.path.join(outPath, outName)
    else:
        fullPath = None
    return fullPath


def fileOpenDlg(tryFilePath="",
                tryFileName="",
                prompt=_translate("Select file(s) to open"),
                allowed=None):
    """A simple dialogue allowing read access to the file system.

    :parameters:
        tryFilePath: string
            default file path on which to open the dialog
        tryFileName: string
            default file name, as suggested file
        prompt: string (default "Select file to open")
            can be set to custom prompts
        allowed: string (available since v1.62.01)
            a string to specify file filters.
            e.g. "BMP files (*.bmp)|*.bmp|GIF files (*.gif)|*.gif"
            See http://www.wxpython.org/docs/api/wx.FileDialog-class.html
            for further details

    If tryFilePath or tryFileName are empty or invalid then
    current path and empty names are used to start search.

    If user cancels, then None is returned.
    """
    if allowed is None:
        allowed = ("PsychoPy Data (*.psydat)|*.psydat|"
                   "txt (*.txt,*.dlm,*.csv)|*.txt;*.dlm;*.csv|"
                   "pickled files (*.pickle, *.pkl)|*.pickle|"
                   "shelved files (*.shelf)|*.shelf|"
                   "All files (*.*)|*.*")
    global app  # avoid recreating for every gui
    app = ensureWxApp()
    dlg = wx.FileDialog(None, prompt, tryFilePath, tryFileName, allowed,
                        wx.FD_OPEN | wx.FD_FILE_MUST_EXIST | wx.FD_MULTIPLE)
    if dlg.ShowModal() == OK:
        # get names of images and their directory
        fullPaths = dlg.GetPaths()
    else:
        fullPaths = None
    dlg.Destroy()
    return fullPaths



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from psychopy import visual, core, event, logging
import numpy as np

logging.console.setLevel(logging.EXP)
c = core.Clock()

from psychopy.visual.textbox2 import TextBox2, allFonts

win = visual.Window([800, 800], monitor='testMonitor')
logging.exp("{:.3f}: created window".format(c.getTime()))

psychopyInfo = u"<b>PsychoPy</b> is an <i>open-source</i> Python application allowing you to run a supercali-fragilisticexpeilidocious wide range of neuroscience, psychology and psychophysics experiments. It’s a free, powerful alternative to Presentation™ or e-Prime™, written in Python (a free alternative to Matlab™ g)."

# preload some chars into a font to see how long it takes
fontSize = 16
arial = allFonts.getFont("Arial", fontSize)
logging.exp("{:.3f}: created font".format(c.getTime()))
nChars = 256
arial.preload(nChars)  # or set to preload specific string of chars
logging.exp("{:.3f}: preloaded {} chars".format(c.getTime(), nChars))

txt1 = TextBox2(win, text="Type here, it's toptastic", font='Times',
                color='black', colorSpace='named', 
                pos=(0, 0.4), letterHeight=0.05, units='height',
                size=[0.8, 0.2],
                anchor='center-top',
                borderColor='lightgrey',
                fillColor='slategrey',
                editable=True)

txt2 = TextBox2(win, text=psychopyInfo, font='Arial',
            pos=(0, -5), anchor='middle', size=(20, None), units='cm',
            lineSpacing=1.1,
            letterHeight=1.,
            color='LightGrey', borderColor='Moccasin', fillColor=None,
            editable=True)

txt3 = TextBox2(win, text='Good for non-editable text (Esc to quit)',
            font='Arial',
            borderColor=None, fillColor=None,
            pos=(-0.5,-0.5), units='height', anchor='bottom-left',
            letterHeight=0.02,
            editable=False)

txt1.autoDraw=True
txt2.autoDraw=True
txt3.autoDraw=True

clock = core.Clock()
t=0
while t<30:
    t= clock.getTime()

    txt2.pos = (0.2*np.sin(t), 0.2*np.cos(t))
            
    if 'escape' in event.getKeys():
        core.quit()
    
    win.flip()

logging.flush()




#!/usr/bin/env python
#  -*- coding: utf-8 -*-

# Part of the psychopy.iohub library.
# Copyright (C) 2012-2016 iSolver Software Solutions
# Distributed under the terms of the GNU General Public License (GPL).
from __future__ import division, absolute_import, print_function

from builtins import zip
from builtins import object
import collections
import copy
import os
from collections import deque
from operator import itemgetter

import numpy as np

from .. import _pkgroot
from .computer import Computer
from ..errors import print2err, printExceptionDetailsToStdErr
from ..util import convertCamelToSnake
from future.utils import with_metaclass

class ioDeviceError(Exception):

    def __init__(self, device, msg):
        Exception.__init__(self, msg)
        self.device = device
        self.msg = msg

    def __str__(self):
        return repr(self)

    def __repr__(self):
        return 'ioDeviceError:\n\tMsg: {0:>s}\n\tDevice: {1}\n'.format(
            self.msg, repr(self.device))


class ioObjectMetaClass(type):

    def __new__(meta, name, bases, dct):
        return type.__new__(meta, name, bases, dct)

    def __init__(cls, name, bases, dct):
        type.__init__(cls, name, bases, dct)

        if '_newDataTypes' not in dct:
            cls._newDataTypes = []

        if '_baseDataTypes' not in dct:
            parent = cls._findDeviceParent(bases)
            if parent:
                cls._baseDataTypes = parent._dataType
            else:
                cls._baseDataTypes = []

        cls._dataType = cls._baseDataTypes + cls._newDataTypes
        cls.CLASS_ATTRIBUTE_NAMES = [e[0] for e in cls._dataType]
        cls.NUMPY_DTYPE = np.dtype(cls._dataType)

        if len(cls.__subclasses__()) == 0 and 'DeviceEvent' in [
                c.__name__ for c in cls.mro()]:
            cls.namedTupleClass = collections.namedtuple(
                name + 'NT', cls.CLASS_ATTRIBUTE_NAMES)

    def _findDeviceParent(cls, bases):
        parent = None
        if len(bases) == 1:
            parent = bases[0]
        else:
            for p in bases:
                if 'Device' in p.__name__:
                    parent = p
                    break
        if parent is None or 'object' in parent.__name__:
            return None
        return parent


class ioObject(with_metaclass(ioObjectMetaClass, object)):
    """The ioObject class is the base class for all ioHub Device and
    DeviceEvent classes.

    Any ioHub Device or DeviceEvent class (i.e devices like Keyboard
    Device, Mouse Device, etc; and device events like Message,
    KeyboardPressEvent, MouseMoveEvent, etc.) also include the methods
    and attributes of this class.

    """
    __slots__ = ['_attribute_values', ]

    def __init__(self, *args, **kwargs):
        self._attribute_values = []

        if len(args) > 0:
            for i, n in enumerate(self.CLASS_ATTRIBUTE_NAMES):
                setattr(self, n, args[i])
                self._attribute_values.append(args[i])

        elif len(kwargs) > 0:
            for key in self.CLASS_ATTRIBUTE_NAMES:
                value = kwargs[key]
                setattr(self, key, value)
                self._attribute_values.append(value)

    def _asDict(self):
        """Return the ioObject in dictionary format, with keys as the
        ioObject's attribute names, and dictionary values equal to the
        attribute values.

        Return (dict): dictionary of ioObjects attribute_name, attributes values.

        """
        return dict(list(zip(self.CLASS_ATTRIBUTE_NAMES, self._attribute_values)))

    def _asList(self):
        """Return the ioObject in list format, which is a 1D list of the
        ioObject's attribute values, in the order the ioObject expects them if
        passed to a class constructor.

        Return (list): 1D list of ioObjects _attribute_values

        """
        return self._attribute_values

    def _asNumpyArray(self):
        """Return the ioObject as a numpy array, with the array values being
        equal to what would be returned by the asList() method, and the array
        cell data types being specified by NUMPY_DTYPE class constant.

        Return (numpy.array): numpy array representation of object.

        """
        return np.array([tuple(self._attribute_values), ], self.NUMPY_DTYPE)

    def _getRPCInterface(self):
        rpcList = []
        dlist = dir(self)
        for d in dlist:
            if d[0] != '_' and d not in ['asNumpyArray', ]:
                if callable(getattr(self, d)):
                    rpcList.append(d)
        return rpcList

########### Base Abstract Device that all other Devices inherit from ##########


class Device(ioObject):
    """The Device class is the base class for all ioHub Device types.

    Any ioHub Device class (i.e Keyboard, Mouse, etc) also include the
    methods and attributes of this class.

    """
    DEVICE_USER_LABEL_INDEX = 0
    DEVICE_NUMBER_INDEX = 1
    DEVICE_MANUFACTURER_NAME_INDEX = 2
    DEVICE_MODEL_NAME_INDEX = 3
    DEVICE_MODEL_NUMBER_INDEX = 4
    DEVICE_SOFTWARE_VERSION_INDEX = 5
    DEVICE_HARDWARE_VERSION_INDEX = 6
    DEVICE_FIRMWARE_VERSION_INDEX = 7
    DEVICE_SERIAL_NUMBER_INDEX = 8
    DEVICE_BUFFER_LENGTH_INDEX = 9
    DEVICE_MAX_ATTRIBUTE_INDEX = 9

    # Multiplier to use to convert this devices event time stamps to sec format.
    # This is set by the author of the device class or interface
    # implementation.
    DEVICE_TIMEBASE_TO_SEC = 1.0

    _baseDataTypes = ioObject._baseDataTypes
    _newDataTypes = [
        # The name given to this device instance. User Defined. Should be
        ('name', '|S24'),
        # unique within all devices of the same type_id for a given experiment.
        # For devices that support multiple connected to the computer at once,
        # with some devices the device_number can be used to select which
        # device ot use.
        ('device_number', np.uint8),
        # The name of the manufacturer for the device being used.
        ('manufacturer_name', '|S64'),
        # The string name of the device model being used. Some devices support
        # different models.
        ('model_name', '|S32'),
        # The device model number being used. Some devices support different
        # models.
        ('model_number', '|S32'),
        # Used to optionally store the devices software / API version being
        # used by the ioHub Device
        ('software_version', '|S8'),
        # Used to optionally store the devices hardware version
        ('hardware_version', '|S8'),
        # Used to optionally store the devices firmware
        ('firmware_version', '|S8'),
        # The serial number for the device being used. Serial numbers 'should'
        # be unique across all devices of the same brand and model.
        ('serial_number', '|S32'),
        # The serial number for the device being used. Serial numbers 'should'
        # be unique across all devices of the same brand and model.
        ('manufacture_date', '|S10'),
        # The maximum size of the device level event buffer for this
        ('event_buffer_length', np.uint16)
        # device instance. If the buffer becomes full, when a new event
        # is added, the oldest event in the buffer is removed.
    ]

    EVENT_CLASS_NAMES = []

    _next_event_id = 1
    _display_device = None
    _iohub_server = None
    next_filter_id = 1
    DEVICE_TYPE_ID = None
    DEVICE_TYPE_STRING = None

    # _hw_interface_status constants
    HW_STAT_UNDEFINED = u"HW_STAT_UNDEFINED"
    HW_STAT_NOT_INITIALIZED = u"HW_NOT_INITIALIZED"
    HW_STAT_ERROR = u"HW_ERROR"
    HW_STAT_OK = u"HW_OK"

    __slots__ = [e[0] for e in _newDataTypes] + ['_hw_interface_status',
                                                 '_hw_error_str',
                                                 '_native_event_buffer',
                                                 '_event_listeners',
                                                 '_iohub_event_buffer',
                                                 '_last_poll_time',
                                                 '_last_callback_time',
                                                 '_is_reporting_events',
                                                 '_configuration',
                                                 'monitor_event_types',
                                                 '_filters']

    def __init__(self, *args, **kwargs):
        #: The user defined name given to this device instance. A device name must be
        #: unique for all devices of the same type_id in a given experiment.
        self.name = None

        #: For device classes that support having multiple of the same type
        #: being monitored by the ioHub Process at the same time (for example XInput gamepads),
        #: device_number is used to indicate which of the connected devices of that
        #: type a given ioHub Device instance should connect to.
        self.device_number = None

        #: The maximum size ( in event objects ) of the device level event buffer for this
        #: device instance. If the buffer becomes full, when a new event
        #: is added, the oldest event in the buffer is removed.
        self.event_buffer_length = None

        #: A list of event class names that can be generated by this device type
        #: and should be monitored and reported by the ioHub Process.
        self.monitor_event_types = None

        #: The name of the manufacturer of the device.
        self.manufacturer_name = None

        #: The model of this Device subclasses instance. Some Device types
        #: explicitedly support different models of the device and use different
        #: logic in the ioHub Device implementation based on the model_name given.
        self.model_name = None

        #: Model number can be optionally used to hold the specific model number
        #: specified on the device.
        self.model_number = None

        #: The software version attribute can optionally be used to store the
        #: devices software / API version being used by the ioHub Device
        self.software_version = None

        #: The hardware version attribute can optionally be used to store the
        #: physical devices hardware version / revision.
        self.hardware_version = None

        #: The firmware version attribute can optionally be used to store the
        #: physical devices hardware version / revision.
        self.firmware_version = None

        #: The unique serial number of the specific device instance being used,
        #: if applicable.
        self.serial_number = None

        #: The manufactured date of the specific device instance being used,
        #: if applicable.(Use DD-MM-YYYY string format.)
        self.manufacture_date = None

        ioObject.__init__(self, *args, **kwargs)

        self._is_reporting_events = kwargs.get('auto_report_events', False)
        self._iohub_event_buffer = dict()
        self._event_listeners = dict()
        self._configuration = kwargs
        self._last_poll_time = 0
        self._last_callback_time = 0
        self._native_event_buffer = deque(maxlen=self.event_buffer_length)
        self._filters = dict()
        self._hw_interface_status = self.HW_STAT_UNDEFINED
        self._hw_error_str = u''

    @staticmethod
    def _getNextEventID():
        n = Device._next_event_id
        Device._next_event_id += 1
        return n

    def getConfiguration(self):
        """Retrieve the configuration settings information used to create the
        device instance. This will the default settings for the device, found
        in iohub.devices.<device_name>.default_<device_name>.yaml, updated
        with any device settings provided via launchHubServer(...).

        Changing any values in the returned dictionary has no effect
        on the device state.

        Args:
            None

        Returns:
            (dict): The dictionary of the device configuration settings used 
            to create the device.

        """
        return self._configuration

    def getEvents(self, *args, **kwargs):
        """Retrieve any DeviceEvents that have occurred since the last call to
        the device's getEvents() or clearEvents() methods.

        Note that calling getEvents() at a device level does not change the Global Event Buffer's
        contents.

        Args:
            event_type_id (int): If specified, provides the ioHub DeviceEvent ID for which events should be returned for.  Events that have occurred but do not match the event ID specified are ignored. Event type ID's can be accessed via the EventConstants class; all available event types are class attributes of EventConstants.

            clearEvents (int): Can be used to indicate if the events being returned should also be removed from the device event buffer. True (the default) indicates to remove events being returned. False results in events being left in the device event buffer.

            asType (str): Optional kwarg giving the object type to return events as. Valid values are 'namedtuple' (the default), 'dict', 'list', or 'object'.

        Returns:
            (list): New events that the ioHub has received since the last getEvents() or clearEvents() call to the device. Events are ordered by the ioHub time of each event, older event at index 0. The event object type is determined by the asType parameter passed to the method. By default a namedtuple object is returned for each event.

        """
        self._iohub_server.processDeviceEvents()
        eventTypeID = None
        clearEvents = True
        if len(args) == 1:
            eventTypeID = args[0]
        elif len(args) == 2:
            eventTypeID = args[0]
            clearEvents = args[1]

        if eventTypeID is None:
            eventTypeID = kwargs.get('event_type_id', None)
            if eventTypeID is None:
                eventTypeID = kwargs.get('event_type', None)
        clearEvents = kwargs.get('clearEvents', True)

        filter_id = kwargs.get('filter_id', None)

        currentEvents = []
        if eventTypeID:
            currentEvents = list(self._iohub_event_buffer.get(eventTypeID, []))

            if filter_id:
                currentEvents = [
                    e for e in currentEvents if e[
                        DeviceEvent.EVENT_FILTER_ID_INDEX] == filter_id]

            if clearEvents is True and len(currentEvents) > 0:
                self.clearEvents(
                    eventTypeID,
                    filter_id=filter_id,
                    call_proc_events=False)
        else:
            if filter_id:
                [currentEvents.extend([fe for fe in l if fe[
                                      DeviceEvent.EVENT_FILTER_ID_INDEX] == filter_id]) for l in list(self._iohub_event_buffer.values())]
            else:
                [currentEvents.extend(l)
                 for l in list(self._iohub_event_buffer.values())]

            if clearEvents is True and len(currentEvents) > 0:
                self.clearEvents(filter_id=filter_id, call_proc_events=False)

        if len(currentEvents) > 0:
            currentEvents = sorted(
                currentEvents, key=itemgetter(
                    DeviceEvent.EVENT_HUB_TIME_INDEX))
        return currentEvents

    def clearEvents(
            self,
            event_type=None,
            filter_id=None,
            call_proc_events=True):
        """Clears any DeviceEvents that have occurred since the last call to
        the device's getEvents(), or clearEvents() methods.

        Note that calling clearEvents() at the device level only clears the
        given device's event buffer. The ioHub Process's Global Event Buffer
        is unchanged.

        Args:
            None

        Returns:
            None

        """
        if call_proc_events:
            self._iohub_server.processDeviceEvents()

        if event_type:
            if filter_id:
                event_que = self._iohub_event_buffer[event_type]
                newque = deque([e for e in event_que if e[
                               DeviceEvent.EVENT_FILTER_ID_INDEX] != filter_id], maxlen=self.event_buffer_length)
                self._iohub_event_buffer[event_type] = newque
            else:
                self._iohub_event_buffer.setdefault(
                    event_type, deque(maxlen=self.event_buffer_length)).clear()
        else:
            if filter_id:
                for event_type, event_deque in list(self._iohub_event_buffer.items()):
                    newque = deque([e for e in event_deque if e[
                                   DeviceEvent.EVENT_FILTER_ID_INDEX] != filter_id], maxlen=self.event_buffer_length)
                    self._iohub_event_buffer[event_type] = newque
            else:
                self._iohub_event_buffer.clear()

    def enableEventReporting(self, enabled=True):
        """
        Specifies if the device should be reporting events to the ioHub Process
        (enabled=True) or whether the device should stop reporting events to the
        ioHub Process (enabled=False).


        Args:
            enabled (bool):  True (default) == Start to report device events to the ioHub Process. False == Stop Reporting Events to the ioHub Process. Most Device types automatically start sending events to the ioHUb Process, however some devices like the EyeTracker and AnlogInput device's do not. The setting to control this behavior is 'auto_report_events'

        Returns:
            bool: The current reporting state.
        """
        self.clearEvents()
        self._is_reporting_events = enabled
        return self._is_reporting_events

    def isReportingEvents(self):
        """Returns whether a Device is currently reporting events to the ioHub
        Process.

        Args: None

        Returns:
            (bool): Current reporting state.

        """
        return self._is_reporting_events

    def _setHardwareInterfaceStatus(self, status, error_msg=u''):
        if status is True:
            self._hw_interface_status = self.HW_STAT_OK
            self._hw_error_str = u""
        elif status is False:
            self._hw_interface_status = self.HW_STAT_ERROR
            self._hw_error_str = error_msg
        else:
            self._hw_interface_status = status
            self._hw_error_str = error_msg

    def getLastInterfaceErrorString(self):
        return self._hw_error_str

    def getInterfaceStatus(self):
        return self._hw_interface_status

    def addFilter(self, filter_file_path, filter_class_name, kwargs):
        """Take the filter_file_path and add the filters module dir to sys.path
        if it does not already exist.

        Then import the filter module (file) class based on filter_class_name.
        Create a filter instance, and add it to the _filters dict:

        self._filters[filter_file_path+'.'+filter_class_name]

        :param filter_path:
        :return:

        """
        try:
            import importlib
            filter_file_path = os.path.normpath(
                os.path.abspath(filter_file_path))
            fdir, ffile = os.path.split(filter_file_path)
            if not ffile.endswith('.py'):
                ffile = ffile + '.py'
            if os.path.isdir(fdir) and os.path.exists(filter_file_path):
                if fdir not in sys.path:
                    sys.path.append(fdir)

                # import module using ffile
                filter_module = importlib.import_module(ffile[:-3])

                # import class filter_class_name
                filter_class = getattr(filter_module, filter_class_name, None)
                if filter_class is None:
                    print2err('Can not create Filter, filter class not found')
                    return -1
                else:
                    # Create instance of class
                    # For now, just use a class level counter.
                    filter_class_instance = filter_class(**kwargs)
                    filter_class_instance._parent_device_type = self.DEVICE_TYPE_ID
                    # Add to filter list for device
                    filter_key = filter_file_path + '.' + filter_class_name
                    filter_class_instance._filter_key = filter_key
                    self._filters[filter_key] = filter_class_instance
                    return filter_class_instance.filter_id

            else:
                print2err('Could not add filter . File not found.')
            return -1
        except Exception:
            printExceptionDetailsToStdErr()
            print2err('ERROR During Add Filter')

    def removeFilter(self, filter_file_path, filter_class_name):
        filter_key = filter_file_path + '.' + filter_class_name
        if filter_key in self._filters:
            del self._filters[filter_key]
            return True
        return False

    def resetFilter(self, filter_file_path, filter_class_name):
        filter_key = filter_file_path + '.' + filter_class_name
        if filter_key in self._filters:
            self._filters[filter_key].reset()
            return True
        return False

    def enableFilters(self, yes=True):
        for f in list(self._filters.values()):
            f.enable = yes

    def _handleEvent(self, e):
        event_type_id = e[DeviceEvent.EVENT_TYPE_ID_INDEX]
        self._iohub_event_buffer.setdefault(
            event_type_id, deque(
                maxlen=self.event_buffer_length)).append(e)

        # Add the event to any filters bound to the device which
        # list wanting the event's type and events filter_id
        input_evt_filter_id = e[DeviceEvent.EVENT_FILTER_ID_INDEX]
        for event_filter in list(self._filters.values()):
            if event_filter.enable is True:
                current_filter_id = event_filter.filter_id
                if current_filter_id != input_evt_filter_id:
                    # stops circular event processing
                    evt_filter_ids = event_filter.input_event_types.get(
                        event_type_id, [])
                    if input_evt_filter_id in evt_filter_ids:
                        event_filter._addInputEvent(copy.deepcopy(e))

    def _getNativeEventBuffer(self):
        return self._native_event_buffer

    def _addNativeEventToBuffer(self, e):
        if self.isReportingEvents():
            self._native_event_buffer.append(e)

    def _addEventListener(self, l, eventTypeIDs):
        for ei in eventTypeIDs:
            self._event_listeners.setdefault(ei, []).append(l)

    def _removeEventListener(self, l):
        for etypelisteners in list(self._event_listeners.values()):
            if l in etypelisteners:
                etypelisteners.remove(l)

    def _getEventListeners(self, forEventType):
        return self._event_listeners.get(forEventType, [])

    def getCurrentDeviceState(self, clear_events=True):
        result_dict = {}
        self._iohub_server.processDeviceEvents()
        events = {str(key): tuple(value)
                  for key, value in list(self._iohub_event_buffer.items())}
        result_dict['events'] = events
        if clear_events:
            self.clearEvents(call_proc_events=False)

        result_dict['reporting_events'] = self._is_reporting_events

        return result_dict

    def resetState(self):
        self.clearEvents()

    def _poll(self):
        """The _poll method is used when an ioHub Device needs to periodically
        check for new events received from the native device / device API.
        Normally this means that the native device interface is using some data
        buffer or queue for new device events until the ioHub Device reads
        them.

        The ioHub Device can *poll* and check for any new events that
        are available, retrieve the new events, and process them
        to create ioHub Events as necessary. Each subclass of ioHub.devives.Device
        that wishes to use event polling **must** override the _poll method
        in the Device classes implementation. The configuration section of the
        iohub_config.yaml for the device **must** also contain the device_timer: interval
        parameter as explained below.

        .. note::
            When an event is created by an ioHub Device, it is represented in
            the form of an ordered list, where the number of elements in the
            list equals the number of public attributes of the event, and the order
            of the element values matches the order that the values would be provided
            to the constructor of the associated DeviceEvent class. This list format
            keeps internal event representation overhead (both in terms of creation
            time and memory footprint) to a minimum. The list event format
            also allows for the most compact representation of the event object
            when being transferred between the ioHub and Experiment processes.

            The ioHub Process can convert these list event representations to
            one of several, user-friendly, object formats ( namedtuple [default], dict, or the correct
            ioHub.devices.DeviceEvent subclass. ) for use within the experiment script.

        If an ioHub Device uses polling to check for new device events, the ioHub
        device configuration must include the following property in the devices
        section of the iohub_config.yaml file for the experiment:

            device_timer:
                interval: sec.msec

        The device_timer.interval preference informs ioHub how often the Device._poll
        method should be called while the Device is running.

        For example:

            device_timer:
                interval: 0.01

        indicates that the Device._poll method should ideally be called every 10 msec
        to check for any new events received by the device hardware interface. The
        correct or optimal value for device_timer.interval depends on the device
        type and the expected rate of device events. For devices that receive events
        rapidly, for example at an average rate of 500 Hz or more, or for devices
        that do not provide native event time stamps (and the ioHub Process must
        time stamp the event) the device_timer.interval should be set to 0.001 (1 msec).

        For devices that receive events at lower rates and have native time stamps
        that are being converted to the ioHub time base, a slower polling rate is
        usually acceptable. A general suggestion would be to set the device_timer.interval
        to be equal to two to four times the expected average event input rate in Hz,
        but not exceeding a device_timer.interval 0.001 seconds (a polling rate of 1000 Hz).
        For example, if a device sends events at an average rate of 60 Hz,
        or once every 16.667 msec, then the polling rate could be set to the
        equivalent of a 120 - 240 Hz. Expressed in sec.msec format,
        as is required for the device_timer.interval setting, this would equal about
        0.008 to 0.004 seconds.

        Of course it would be ideal if every device that polled for events was polling
        at 1000 to 2000 Hz, or 0.001 to 0.0005 msec, however if too many devices
        are requested to poll at such high rates, all will suffer in terms of the
        actual polling rate achieved. In devices with slow event output rates,
        such high polling rates will result in many calls to Device._poll that do
        not find any new events to process, causing extra processing overhead that
        is not needed in many cases.

        Args:
            None

        Returns:
            None

        """
        pass

    def _handleNativeEvent(self, *args, **kwargs):
        """The _handleEvent method can be used by the native device interface
        (implemented by the ioHub Device class) to register new native device
        events by calling this method of the ioHub Device class.

        When a native device interface uses the _handleNativeEvent method it is
        employing an event callback approach to notify the ioHub Process when new
        native device events are available. This is in contrast to devices that use
        a polling method to check for new native device events, which would implement
        the _poll() method instead of this method.

        Generally speaking this method is called by the native device interface
        once for each new event that is available for the ioHub Process. However,
        with good cause, there is no reason why a single call to this
        method could not handle multiple new native device events.

        .. note::
            If using _handleNativeEvent, be sure to remove the device_timer
            property from the devices configuration section of the iohub_config.yaml.

        Any arguments or kwargs passed to this method are determined by the ioHub
        Device implementation and should contain all the information needed to create
        an ioHub Device Event.

        Since any callbacks should take as little time to process as possible,
        a two stage approach is used to turn a native device event into an ioHub
        Device event representation:
            #. This method is called by the native device interface as a callback, providing the necessary information to be able to create an ioHub event. As little processing should be done in this method as possible.
            #. The data passed to this method, along with the time the callback was called, are passed as a tuple to the Device classes _addNativeEventToBuffer method.
            #. During the ioHub Servers event processing routine, any new native events that have been added to the ioHub Server using the _addNativeEventToBuffer method are passed individually to the _getIOHubEventObject method, which must also be implemented by the given Device subclass.
            #. The _getIOHubEventObject method is responsible for the actual conversion of the native event representation to the required ioHub Event representation for the accociated event type.

        Args:
            args(tuple): tuple of non keyword arguements passed to the callback.

        Kwargs:
            kwargs(dict): dict of keyword arguements passed to the callback.

        Returns:
            None

        """
        return False

    def _getIOHubEventObject(self, native_event_data):
        """The _getIOHubEventObject method is called by the ioHub Process to
        convert new native device event objects that have been received to the
        appropriate ioHub Event type representation.

        If the ioHub Device has been implemented to use the _poll() method of checking for
        new events, then this method simply should return what it is passed, and is the
        default implementation for the method.

        If the ioHub Device has been implemented to use the event callback method
        to register new native device events with the ioHub Process, then this method should be
        overwritten by the Device subclass to convert the native event data into
        an appropriate ioHub Event representation. See the implementation of the
        Keyboard or Mouse device classes for an example of such an implementation.

        Args:
            native_event_data: object or tuple of (callback_time, native_event_object)

        Returns:
            tuple: The appropriate ioHub Event type in list form.

        """
        return native_event_data

    def _close(self):
        try:
            self.__class__._iohub_server = None
            self.__class__._display_device = None
        except Exception:
            pass

    def __del__(self):
        self._close()

########### Base Device Event that all other Device Events inherit from ##


class DeviceEvent(ioObject):
    """The DeviceEvent class is the base class for all ioHub DeviceEvent types.

    Any ioHub DeviceEvent class (i.e MouseMoveEvent, MouseScrollEvent,
    MouseButtonPressEvent, KeyboardPressEvent, KeyboardReleaseEvent,
    etc.) also has access to the methods and attributes of the
    DeviceEvent class.

    """
    EVENT_EXPERIMENT_ID_INDEX = 0
    EVENT_SESSION_ID_INDEX = 1
    DEVICE_ID_INDEX = 2
    EVENT_ID_INDEX = 3
    EVENT_TYPE_ID_INDEX = 4
    EVENT_DEVICE_TIME_INDEX = 5
    EVENT_LOGGED_TIME_INDEX = 6
    EVENT_HUB_TIME_INDEX = 7
    EVENT_CONFIDENCE_INTERVAL_INDEX = 8
    EVENT_DELAY_INDEX = 9
    EVENT_FILTER_ID_INDEX = 10
    BASE_EVENT_MAX_ATTRIBUTE_INDEX = EVENT_FILTER_ID_INDEX

    # The Device Class that generates the given type of event.
    PARENT_DEVICE = None

    # The string label for the given DeviceEvent type. Should be usable to get Event type
    #  from ioHub.EventConstants.getName(EVENT_TYPE_STRING), the value of which is the
    # event type id. This is set by the author of the event class
    # implementation.
    EVENT_TYPE_STRING = 'UNDEFINED_EVENT'

    # The type id int for the given DeviceEvent type. Should be one of the int values in
    # ioHub.EventConstants.EVENT_TYPE_ID. This is set by the author of the
    # event class implementation.
    EVENT_TYPE_ID = 0

    _baseDataTypes = ioObject._baseDataTypes
    _newDataTypes = [
        # The ioDataStore experiment ID assigned to the experiment code
        ('experiment_id', np.uint8),
        # specified in the experiment configuration file for the experiment.

        # The ioDataStore session ID assigned to the currently running
        ('session_id', np.uint8),
        # experiment session. Each time the experiment script is run,
        # a new session id is generated for use within the hdf5 file.

        # The unique id assigned to the device that generated the event.
        ('device_id', np.uint8),
        # CUrrrently not used, but will be in the future for device types that
        # support > one instance of that device type to be enabled
        # during an experiment. Currently only one device of a given type
        # can be used in an experiment.

        # The id assigned to the current device event instance. Every device
        ('event_id', np.uint32),
        # event generated by monitored devices during an experiment session is
        # assigned a unique id, starting from 0 for each session, incrementing
        # by +1 for each new event.

        # The type id for the event. This is used to create DeviceEvent objects
        ('type', np.uint8),
        # or dictionary representations of an event based on the data from an
        # event value list.

        # If the device that generates the given device event type also time
        # stamps
        ('device_time', np.float64),
        # events, this field is the time of the event as given by the device,
        # converted to sec.msec-usec for consistency with all other ioHub device times.
        # If the device that generates the given event type does not time stamp
        # events, then the device_time is set to the logged_time for the event.

        # The sec time that the event was 'received' by the ioHub Server
        # Process.
        ('logged_time', np.float64),
        # For devices that poll for events, this is the sec time that the poll
        # method was called for the device and the event was retrieved. For
        # devices that use the event callback, this is the sec time the callback
        # executed and accept the event. Time is in sec.msec-usec

        # Time is in the normalized time base that all events share,
        ('time', np.float64),
        # regardless of device type. Time is calculated differently depending
        # on the device and perhaps event type.
        # Time is what should be used when comparing times of events across
        # different devices. Time is in sec.msec-usec.

        # This property attempts to give a sense of the amount to which
        ('confidence_interval', np.float32),
        # the event time may be off relative to the true time the event
        # occurred. confidence_interval is calculated differently depending
        # on the device and perhaps event types. In general though, the
        # smaller the confidence_interval, the more likely it is that the
        # calculated time of the event is correct. For devices where
        # a realistic confidence_interval can not be calculated,
        # for example if the event device delay is unknown, then a value
        # of -1.0 should be used. Valid confidence_interval values are
        # in sec.msec-usec and will range from 0.000000 sec.msec-usec
        # and higher.

        # The delay of an event is the known (or estimated) delay from when the
        ('delay', np.float32),
        # real world event occurred to when the ioHub received the event for
        # processing. This is often called the real-time end-to-end delay
        # of an event. If the delay for an event can not be reasonably estimated
        # or is not known, a delay of -1.0 is set. Delays are in sec.msec-usec
        # and valid values will range from 0.000000 sec.msec-usec and higher.

        # The filter identifier that the event passed through before being
        # saved.
        ('filter_id', np.int16)
        # If the event did not pass through any filter devices, then the value will be 0.
        # Otherwise, the value is the | combination of the filter set that the
        # event passed through before being made available to the experiment,
        # or saved to the ioDataStore. The filter id can be used to determine
        # which filters an event was handled by, but not the order in which handling was done.
        # Default value is 0.
    ]

    # The name of the hdf5 table used to store events of this type in the ioDataStore pytables file.
    # This is set by the author of the event class implementation.
    IOHUB_DATA_TABLE = None

    __slots__ = [e[0] for e in _newDataTypes]

    def __init__(self, *args, **kwargs):
        #: The ioHub DataStore experiment ID assigned to the experiment that is running when the event is collected.
        #: 0 indicates no experiment has been defined.
        self.experiment_id = None

        #: The ioHub DataStore session ID assigned for the current experiment run.
        #: Each time the experiment script is run, a new session id is generated for use
        #: by the ioHub DataStore within the hdf5 file.
        self.session_id = None

        self.device_id = None

        #: The id assigned to the current event instance. Every device
        #: event generated by the ioHub Process is assigned a unique id,
        #: starting from 0 for each session, incrementing by +1 for each new event.
        self.event_id = None

        #: The type id for the event. This is used to create DeviceEvent objects
        #: or dictionary representations of an event based on the data from an
        #: event value list. Event types are all defined in the
        #: iohub.constants.EventConstants class as class attributes.
        self.type = None

        #: If the device that generates an event type also time stamps
        #: the events, this field is the time of the event as given by the device,
        #: converted to sec.msec-usec for consistency with all other device times.
        #: If the device that generates the event does not time stamp
        #: events, then the device_time is set to the logged_time for the event.
        self.device_time = None

        #: The sec.msec time that the event was 'received' by the ioHub Process.
        #: For devices where the ioHub polls for events, this is the time that the poll
        #: method was called for the device and the event was retrieved. For
        #: devices that use the event callback to inform the ioHub of new events,
        #: this is the time the callback began to be executed. Time is in sec.msec-usec
        self.logged_time = None

        #: The calculated ioHub time is in the normalized time base that all events share,
        #: regardless of device type. Time is calculated differently depending
        #: on the device and perhaps event type.
        #: Time is what should be used when comparing times of events across
        #: different devices or with times given py psychopy.core.getTime(). Time is in sec.msec-usec.
        self.time = None

        #: This property attempts to give a sense of the amount to which
        #: the event time may be off relative to the true time the event
        #: may have become available to the ioHub Process.
        #: confidence_interval is calculated differently depending
        #: on the device and perhaps event types. In general though, the
        #: smaller the confidence_interval, the more accurate the
        #: calculated time of the event will be. For devices where
        #: a meaningful confidence_interval can not be calculated, a value
        #: of 0.0 is used. Valid confidence_interval values are
        #: in sec.msec-usec and will range from 0.000001 sec.msec-usec
        #: and higher.
        self.confidence_interval = None

        #: The delay of an event is the known (or estimated) delay from when the
        #: real world event occurred to when the ioHub received the event for
        #: processing. This is often called the real-time end-to-end delay
        #: of an event. If the delay for an event can not be reasonably estimated
        #: or is not known at all, a delay of 0.0 is set. Delays are in sec.msec-usec
        #: and valid values will range from 0.000001 sec.msec-usec and higher.
        #: the delay of an event is suptracted from the initially determined ioHub
        #: time for the eventso that the event.time attribute reports the actual
        #: event time as accurately as possible.
        self.delay = None

        self.filter_id = None

        ioObject.__init__(self, *args, **kwargs)

    def __cmp__(self, other):
        return self.time - other.time

    @classmethod
    def createEventAsClass(cls, eventValueList):
        kwargs = cls.createEventAsDict(eventValueList)
        return cls(**kwargs)

    @classmethod
    def createEventAsDict(cls, values):
        return dict(list(zip(cls.CLASS_ATTRIBUTE_NAMES, values)))

    # noinspection PyUnresolvedReferences
    @classmethod
    def createEventAsNamedTuple(cls, valueList):
        return cls.namedTupleClass(*valueList)
#
# Import Devices and DeviceEvents
#


import sys


def import_device(module_path, device_class_name):
    module = __import__(module_path, fromlist=["{}".format(device_class_name)])
    device_class = getattr(module, device_class_name)

    setattr(sys.modules[__name__], device_class_name, device_class)

    event_classes = dict()

    for event_class_name in device_class.EVENT_CLASS_NAMES:
        event_constant_string = convertCamelToSnake(
            event_class_name[:-5], False)

        event_module = __import__(module_path, fromlist=[event_class_name])
        event_class = getattr(event_module, event_class_name)

        event_class.DEVICE_PARENT = device_class

        event_classes[event_constant_string] = event_class

        setattr(sys.modules[__name__], event_class_name, event_class)

    return device_class, device_class_name, event_classes

try:
    if getattr(sys.modules[__name__], 'Display', None) is None:
        display_class, device_class_name, event_classes = import_device(
            '%s.devices.display' % (_pkgroot), 'Display')
        setattr(sys.modules[__name__], 'Display', display_class)
except Exception:
    print2err('Warning: display device module could not be imported.')
    printExceptionDetailsToStdErr()



# -*- coding: utf-8 -*-
# Part of the psychopy.iohub library.
# Copyright (C) 2012-2016 iSolver Software Solutions
# Distributed under the terms of the GNU General Public License (GPL).
from __future__ import division, print_function, absolute_import

import ctypes

from . import MouseDevice
from ...constants import EventConstants, MouseConstants
from .. import Computer, Device
from ..keyboard import Keyboard
from ...errors import print2err

currentSec = Computer.getTime

class Mouse(MouseDevice):
    """The Mouse class and related events represent a standard computer mouse
    device and the events a standard mouse can produce.

    Mouse position data is mapped to the coordinate space defined in the
    ioHub configuration file for the Display.

    Examples:

        A. Print all mouse events received for 5 seconds::
    
            from psychopy.iohub import launchHubServer
            from psychopy.core import getTime
            
            # Start the ioHub process. 'io' can now be used during the
            # experiment to access iohub devices and read iohub device events.
            io = launchHubServer()
            
            mouse = io.devices.mouse
                    
            # Check for and print any Mouse events received for 5 seconds.
            stime = getTime()
            while getTime()-stime < 5.0:
                for e in mouse.getEvents():
                    print(e)
            
            # Stop the ioHub Server
            io.quit()
            
        B. Print current mouse position for 5 seconds::
    
            from psychopy.iohub import launchHubServer
            from psychopy.core import getTime
            
            # Start the ioHub process. 'io' can now be used during the
            # experiment to access iohub devices and read iohub device events.
            io = launchHubServer()
            
            mouse = io.devices.mouse
                    
            # Check for and print any Mouse events received for 5 seconds.
            stime = getTime()
            while getTime()-stime < 5.0:
                print(mouse.getPosition())
            
            # Stop the ioHub Server
            io.quit()
            
    """
    WM_MOUSEFIRST = 0x0200
    WM_MOUSEMOVE = 0x0200
    WM_LBUTTONDOWN = 0x0201
    WM_LBUTTONUP = 0x0202
    WM_LBUTTONDBLCLK = 0x0203
    WM_RBUTTONDOWN = 0x0204
    WM_RBUTTONUP = 0x0205
    WM_RBUTTONDBLCLK = 0x0206
    WM_MBUTTONDOWN = 0x0207
    WM_MBUTTONUP = 0x0208
    WM_MBUTTONDBLCLK = 0x0209
    WM_MOUSEWHEEL = 0x020A
    WM_MOUSELAST = 0x020A

    WH_MOUSE = 7
    WH_MOUSE_LL = 14
    WH_MAX = 15

    _mouse_event_mapper = {
        WM_MOUSEMOVE: [
            0,
            EventConstants.MOUSE_MOVE,
            MouseConstants.MOUSE_BUTTON_NONE],
        WM_RBUTTONDOWN: [
            MouseConstants.MOUSE_BUTTON_STATE_PRESSED,
            EventConstants.MOUSE_BUTTON_PRESS,
            MouseConstants.MOUSE_BUTTON_RIGHT],
        WM_MBUTTONDOWN: [
            MouseConstants.MOUSE_BUTTON_STATE_PRESSED,
            EventConstants.MOUSE_BUTTON_PRESS,
            MouseConstants.MOUSE_BUTTON_MIDDLE],
        WM_LBUTTONDOWN: [
            MouseConstants.MOUSE_BUTTON_STATE_PRESSED,
            EventConstants.MOUSE_BUTTON_PRESS,
            MouseConstants.MOUSE_BUTTON_LEFT],
        WM_RBUTTONUP: [
            MouseConstants.MOUSE_BUTTON_STATE_RELEASED,
            EventConstants.MOUSE_BUTTON_RELEASE,
            MouseConstants.MOUSE_BUTTON_RIGHT],
        WM_MBUTTONUP: [
            MouseConstants.MOUSE_BUTTON_STATE_RELEASED,
            EventConstants.MOUSE_BUTTON_RELEASE,
            MouseConstants.MOUSE_BUTTON_MIDDLE],
        WM_LBUTTONUP: [
            MouseConstants.MOUSE_BUTTON_STATE_RELEASED,
            EventConstants.MOUSE_BUTTON_RELEASE,
            MouseConstants.MOUSE_BUTTON_LEFT],
        WM_RBUTTONDBLCLK: [
            MouseConstants.MOUSE_BUTTON_STATE_MULTI_CLICK,
            EventConstants.MOUSE_MULTI_CLICK,
            MouseConstants.MOUSE_BUTTON_RIGHT],
        WM_MBUTTONDBLCLK: [
            MouseConstants.MOUSE_BUTTON_STATE_MULTI_CLICK,
            EventConstants.MOUSE_MULTI_CLICK,
            MouseConstants.MOUSE_BUTTON_MIDDLE],
        WM_LBUTTONDBLCLK: [
            MouseConstants.MOUSE_BUTTON_STATE_MULTI_CLICK,
            EventConstants.MOUSE_MULTI_CLICK,
            MouseConstants.MOUSE_BUTTON_LEFT],
        WM_MOUSEWHEEL: [
            0,
            EventConstants.MOUSE_SCROLL,
            MouseConstants.MOUSE_BUTTON_NONE]}

    slots = ['_user32', '_original_system_cursor_clipping_rect']

    def __init__(self, *args, **kwargs):
        MouseDevice.__init__(self, *args, **kwargs['dconfig'])

        self._user32 = ctypes.windll.user32

        self._original_system_cursor_clipping_rect = ctypes.wintypes.RECT()
        self._user32.GetClipCursor(ctypes.byref(
            self._original_system_cursor_clipping_rect))

    def _initialMousePos(self):
        """If getPosition is called prior to any mouse events being received,
        this method gets the current system cursor pos.
        """
        if self._position is None:
            p = 0.0, 0.0
            mpos = ctypes.wintypes.POINT()
            if self._user32.GetCursorPos(ctypes.byref(mpos)):
                display_index = self.getDisplayIndexForMousePosition(
                    (mpos.x,mpos.y))
    
                if display_index == -1 and self._last_display_index is not None:
                    display_index = self._last_display_index

                if display_index != self._display_device.getIndex():
                    # sys mouse is currently not in psychopy window
                    # so keep pos to window center.
                    display_index = -1
        
                if display_index == -1:
                    self._display_index = self._display_device.getIndex()
                    self._last_display_index = self._display_index
                    wm_pix = self._display_device._displayCoord2Pixel(p[0],
                                                                      p[1],
                                                                      self._display_index)
                    self._nativeSetMousePos(*wm_pix)
                else:
                    p = self._display_device._pixel2DisplayCoord(
                        mpos.x, mpos.y, display_index)
                
                self._position = p
                self._lastPosition = self._position

    def _nativeSetMousePos(self, px, py):
        self._user32.SetCursorPos(int(px), int(py))
        #print2err("_nativeSetMousePos {0}".format((int(px), int(py))))

    def _nativeEventCallback(self, event):
        if self.isReportingEvents():
            logged_time = currentSec()
            report_system_wide_events = self.getConfiguration().get(
                'report_system_wide_events', True)
            pyglet_window_hnds = self._iohub_server._pyglet_window_hnds
            if event.Window in pyglet_window_hnds:
                pass
            elif len(pyglet_window_hnds) > 0 and report_system_wide_events is False:
                return True
            self._scrollPositionY += event.Wheel
            event.WheelAbsolute = self._scrollPositionY

            display_index = self.getDisplayIndexForMousePosition(
                event.Position)

            if display_index == -1:
                if self._last_display_index is not None:
                    display_index = self._last_display_index
                else:
                    return True

            mx, my = event.Position
            event.DisplayIndex = display_index

            #print2err("handleMouseEvt: pix pos={}, display_index: {}".format((int(mx), int(my)),display_index))

            p = self._display_device._pixel2DisplayCoord(
                mx, my, event.DisplayIndex)

            event.Position = p

            self._lastPosition = self._position
            self._position = event.Position

            self._last_display_index = self._display_index
            self._display_index = display_index

            bstate, etype, bnum = self._mouse_event_mapper[event.Message]
            if bnum is not MouseConstants.MOUSE_BUTTON_NONE:
                self.activeButtons[bnum] = int(
                    bstate == MouseConstants.MOUSE_BUTTON_STATE_PRESSED)

            abuttonSum = 0
            for k, v in self.activeButtons.items():
                abuttonSum += k * v

            event.ActiveButtons = abuttonSum

            self._addNativeEventToBuffer((logged_time, event))

            self._last_callback_time = logged_time

        # pyHook require the callback to return True to inform the windows
        # low level hook functionality to pass the event on.
        return True

    def _getIOHubEventObject(self, native_event_data):
        logged_time, event = native_event_data
        p = event.Position
        px = p[0]
        py = p[1]
        bstate, etype, bnum = self._mouse_event_mapper[event.Message]
        if event.Message == self.WM_MOUSEMOVE and event.ActiveButtons > 0:
            etype = EventConstants.MOUSE_DRAG
        confidence_interval = 0.0
        delay = 0.0
        device_time = event.Time / 1000.0  # convert to sec
        hubTime = logged_time
        r = [0,
             0,
             0,  # device id
             Device._getNextEventID(),
             etype,
             device_time,
             logged_time,
             hubTime,
             confidence_interval,
             delay,
             0,
             event.DisplayIndex,
             bstate,
             bnum,
             event.ActiveButtons,
             px,
             py,
             0,  # scroll_dx not supported
             0,  # scroll_x not supported
             event.Wheel,
             event.WheelAbsolute,
             Keyboard._modifier_value,
             event.Window]
        return r

    def __del__(self):
        self._user32.ClipCursor(
            ctypes.byref(
                self._original_system_cursor_clipping_rect))
        MouseDevice.__del__(self)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

'''Deprecated (as of version 1.74.00):
please use the :class:`~psychopy.visual.GratingStim`
or the :class:`~psychopy.visual.ImageStim` classes.'''

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2021 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

from __future__ import absolute_import, print_function

from psychopy.visual.grating import GratingStim


class PatchStim(GratingStim):

    def __init__(self, *args, **kwargs):
        """
        Deprecated (as of version 1.74.00):
        please use the :class:`~psychopy.visual.GratingStim`
        or the :class:`~psychopy.visual.ImageStim` classes.

        The GratingStim has identical abilities to the PatchStim
        (but possibly different initial values)
        whereas the ImageStim is designed to be use for non-cyclic images
        (photographs, not gratings).
        """
        super(PatchStim, self).__init__(*args, **kwargs)
        self.setImage = self.setTex



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2021 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

import wx


class SyncStatusPanel(wx.Panel):
    def __init__(self, parent, id, *args, **kwargs):
        # init super classes
        wx.Panel.__init__(self, parent, id, *args, **kwargs)
        # set self properties
        self.parent = parent
        self.infoStream = InfoStream(self, -1, size=(250, 150))
        # self.progBar = wx.Gauge(self, -1, range=1, size=(200, -1))

        self.mainSizer = wx.BoxSizer(wx.VERTICAL)
        self.mainSizer.Add(self.infoStream, 1, wx.ALL | wx.CENTER | wx.EXPAND,
                           border=10)
        # self.mainSizer.Add(self.progBar, 1, wx.ALL | wx.CENTER, border=10)

        self.SetAutoLayout(True)
        self.SetSizerAndFit(self.mainSizer)
        self.Layout()

    def setStatus(self, status):
        self.infoStream.SetValue(status)
        self.Refresh()
        self.Layout()
        wx.Yield()

    def statusAppend(self, newText):
        text = self.infoStream.GetValue() + newText
        self.setStatus(text)


class InfoStream(wx.TextCtrl):
    def __init__(self, parent, id, size,
                 value="Synchronising...",
                 style=wx.TE_READONLY | wx.TE_MULTILINE):
        wx.TextCtrl.__init__(self, parent, id,
                             size=size, value=value, style=style)

    def clear(self):
        self.SetValue("")

    def write(self, text):
        if type(text) == bytes:
            text = text.decode('utf-8')
        self.SetValue(self.GetValue() + text)
        wx.Yield()



from psychopy import visual, core, prefs
from pathlib import Path
from math import sin, pi, cos

win = visual.Window(size=(128,128), units='pix')

resourcesFolder = Path(prefs.paths['resources'])
psychopyIcon = resourcesFolder / "psychopy.png"
img = visual.ImageStim(win, psychopyIcon)
img.autoDraw = True

# grow and spin
for frameN in range(200):
    if frameN < 50:
        h = frameN*2
    img.size = (cos(frameN/50*pi)*h, h)
    win.flip()
    win.getMovieFrame()

# fade to grey
for frameN in range(5):
    img.contrast -= 0.2
    win.flip()
    win.getMovieFrame()
    
win.saveMovieFrames('myMov.mp4', fps=30, codec='libx264', clearFrames=False)
#win.saveMovieFrames('myMov.png')
#win.saveMovieFrames('myMov.gif')


#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2021 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

from __future__ import absolute_import, print_function
from builtins import super  # provides Py3-style super() using python-future

from os import path
from psychopy.experiment.components import BaseComponent, Param, _translate
from psychopy.localization import _localized as __localized
_localized = __localized.copy()

# the absolute path to the folder containing this path
thisFolder = path.abspath(path.dirname(__file__))
iconFile = path.join(thisFolder, 'eyetracker.png')
tooltip = _translate('Eyetracker: use one of several eyetrackers to follow '
                     'gaze')


class EyetrackerComponent(BaseComponent):
    """A class for using one of several eyetrackers to follow gaze"""
    categories = ['Responses']

    def __init__(self, exp, parentName, name='eyes',
                 startType='time (s)', startVal=0.0,
                 stopType='duration (s)', stopVal=1.0,
                 startEstim='', durationEstim='',
                 save='final', configFile='myTracker.yaml'):
        self.type = 'Eyetracker'
        self.url = "https://www.psychopy.org/builder/components/eyetracker.html"
        self.parentName = parentName
        self.exp = exp  # so we can access the experiment if necess
        self.exp.requirePsychopyLibs(['iohub'])
        # params
        self.params = {}
        self.order = ['Config file']  # first param after the name

        # useful params for the eyetracker - keep to a minimum if possible! ;-)
        self.params['Config file'] = Param(
            configFile, valType='str', categ='Hardware',
            hint=_translate("Config file for eyetracker parameters"))

        msg = _translate(
            "How often should the eyetracker state (x,y,"
            "pupilsize...) be stored? On every video frame, every click "
            "or just at the end of the Routine?")
        self.params['saveState'] = Param(
            save, valType='str', inputType="choice", categ='Data',
            allowedVals=['final', 'every frame', 'never'],
            hint=msg,
            label="Save eyetracker state")

    def writePreWindowCode(self, buff):
        buff.writeIndented("#%(name)s: do calibration\n" % self.params)

        # these might move to a more general place later, when we're always
        # planning on having iohub running
        code = ("io_config = iohub.load(open(%('Config file')s,'r'), "
                "Loader=iohub.Loader)\n"
                "io = iohub.ioHubConnection(io_config)\n"
                "eyetracker = io.getDevice('tracker')\n"
                "eyetracker.runSetupProcedure()\n")
        buff.writeIndentedLines(code % self.params)

    def writeInitCode(self, buff):
        pass  # do we need anything after window creation but before run starts?

    def writeRoutineStartCode(self, buff):
        """Write the code that will be called at the start of the routine
        """
        # create some lists to store recorded values positions and events if we
        # need more than one
        code = "# setup some python lists for storing info about the %(name)s\n" % self.params
        buff.writeIndented(code)
        # a list of vals for each val, rather than a scalar
        if self.params['saveState'].val in ['every frame', 'on click']:
            code = ("%(name)s.x = []\n"
                    "%(name)s.y = []\n"
                    # is this common or is Jon making it up?!
                    "%(name)s.pupil = []\n")
            buff.writeIndentedLines(code % self.params)

    def writeFrameCode(self, buff):
        """Write the code that will be called every frame
        """
        buff.writeIndented("# *%s* updates\n" % self.params['name'])

        # test for whether we're just starting to record
        # writes an if statement to determine whether to draw etc
        self.writeStartTestCode(buff)
        code = ("%(name)s.status = STARTED\n"
                "# clear events and start tracking\n"
                "io.clearEvents('all')\n"
                "%(name)s.setRecordingState(True)\n")
        buff.writeIndentedLines(code % self.params)

        # to get out of the if statement
        buff.setIndentLevel(-1, relative=True)

        # test for stop (only if there was some setting for duration or stop)
        if self.params['stopVal'].val not in ['', None, -1, 'None']:
            # writes an if statement to determine whether to draw etc
            self.writeStopTestCode(buff)

            code = ("%(name)s.status = FINISHED\n"
                    "%(name)s.setRecordingState(False)\n")
            buff.writeIndentedLines(code % self.params)

            # to get out of the if statement
            buff.setIndentLevel(-2, relative=True)

        # if STARTED and not FINISHED!
        code = "if %(name)s.status == STARTED:  # only update if started and not finished!\n" % self.params
        buff.writeIndented(code)

        buff.setIndentLevel(1, relative=True)  # to get out of the if statement
        dedentAtEnd = 1  # keep track of how far to dedent later

        code = ("%(name)s.x, %(name)s.y = eyetracker.getPosition()\n"
                "%(name)s.pupil = eyetracker.getPupilSize()\n")
        buff.writeIndentedLines(code % self.params)

        # actual each-frame checks
        if self.params['saveState'].val in ['every frame']:
            code = ("x, y = eyetracker.getPosition()\n"
                    "%(name)s.x.append(x)\n"
                    "%(name)s.y.append(y)\n"
                    "%(name)s.pupil.append(eyetracker.getPupilSize())\n")
            buff.writeIndented(code % self.params)

        # dedent
        # 'if' statement of the time test and button check
        buff.setIndentLevel(-dedentAtEnd, relative=True)

    def writeRoutineEndCode(self, buff):
        # some shortcuts
        name = self.params['name'].val
        # do this because the param itself is not a string!
        store = self.params['saveState'].val
        # check if we're in a loop (so saving is possible)
        if len(self.exp.flow._loopList):
            currLoop = self.exp.flow._loopList[-1]  # last (outer-most) loop
        else:
            currLoop = None

        # if store=='final' then update value
        if store == 'final' and currLoop != None:
            code = ("# get info about the %(name)s\n"
                    "%(name)s.x, %(name)s.y = eyetracker.getPosition()\n"
                    "x, y = %(name)s.getPos()\n"
                    "%(name)s.pupil = eyetracker.getPupilSize()\n")
            buff.writeIndentedLines(code % self.params)

        # then push to psychopy data file if store='final','every frame'
        if store != 'never' and currLoop != None:
            buff.writeIndented("# save %(name)s data\n" % self.params)
            for property in ['x', 'y', 'pupil']:
                buff.writeIndented(
                    "%s.addData('%s.%s', %s.%s)\n" %
                    (currLoop.params['name'], name, property, name, property))

        # make sure eyetracking stops recording (in case it hasn't stopped
        # already)
        buff.writeIndented("eyetracker.setRecordingState(False)\n")

        # get parent to write code too (e.g. store onset/offset times)
        super().writeRoutineEndCode(buff)

    def writeExperimentEndCode(self, buff):
        buff.writeIndented("eyetracker.setConnectionState(False)\n")
        # in future this should be done generally, not by the eyetracker
        buff.writeIndented("io.quit()\n")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""Writes the current version, build platform etc.
"""

from __future__ import absolute_import, print_function
from setuptools.config import read_configuration
import os
import platform
import subprocess
from psychopy.constants import PY3
from pathlib import Path

root = Path(__file__).parent.parent
# import versioneer
# get version from file
with open('version') as f:
    version = f.read().strip()


def createInitFile(dist=None, version=None, sha=None):
    """Create psychopy/__init__.py

    :param:`dist` can be:
        None:
            writes __version__
        'sdist':
            for python setup.py sdist - writes git id (__git_sha__)
        'bdist':
            for python setup.py bdist - writes git id (__git_sha__)
            and __build_platform__
    """
    # get default values if None
    if version is None:
        with open(str(root/'version')) as f:
            version = f.read().strip()
    if sha is None:
        sha = _getGitShaString(dist)
    platformStr = _getPlatformString(dist)

    metadata = read_configuration('setup.cfg')['metadata']
    infoDict = {'version': version,
                'author': metadata['author'],
                'author_email': metadata['author_email'],
                'maintainer_email': metadata['maintainer_email'],
                'url': metadata['url'],
                'download_url': metadata['download_url'],
                'license': metadata['license'],
                'shaStr': sha,
                'platform': platformStr}

    # write it
    with open(str(root/'psychopy/__init__.py'), 'w') as f:
        outStr = template.format(**infoDict)
        f.write(outStr)
    print('wrote init for ', version, sha)
    # and return it
    return outStr


template = """#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2021 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

# --------------------------------------------------------------------------
# This file is automatically generated during build (do not edit directly).
# --------------------------------------------------------------------------

import os
import sys

__version__ = '{version}'
__license__ = '{license}'
__author__ = '{author}'
__author_email__ = '{author_email}'
__maintainer_email__ = '{maintainer_email}'
__url__ = '{url}'
__download_url__ = '{download_url}'
__git_sha__ = '{shaStr}'
__build_platform__ = '{platform}'

__all__ = ["gui", "misc", "visual", "core",
           "event", "data", "sound", "microphone"]

# for developers the following allows access to the current git sha from
# their repository
if __git_sha__ == 'n/a':
    from subprocess import check_output, PIPE
    # see if we're in a git repo and fetch from there
    try:
        thisFileLoc = os.path.split(__file__)[0]
        output = check_output(['git', 'rev-parse', '--short', 'HEAD'],
                              cwd=thisFileLoc, stderr=PIPE)
    except Exception:
        output = False
    if output:
        __git_sha__ = output.strip()  # remove final linefeed

# update preferences and the user paths
if 'installing' not in locals():
    from psychopy.preferences import prefs
    for pathName in prefs.general['paths']:
        sys.path.append(pathName)

    from psychopy.tools.versionchooser import useVersion, ensureMinimal

# import readline here to get around an issue with sounddevice
# issues GH-2230 GH-2344 GH-2662
try:
    import readline
except ImportError:
    pass  # all that will happen is the stderr/stdout might get redirected

"""


def _getGitShaString(dist=None, sha=None):
    """If generic==True then returns empty __git_sha__ string
    """
    shaStr = 'n/a'
    if dist is not None:
        proc = subprocess.Popen('git rev-parse --short HEAD',
                                stdout=subprocess.PIPE,
                                stderr=subprocess.PIPE,
                                cwd='.', shell=True)
        repo_commit, _ = proc.communicate()
        if PY3:
            repo_commit = repo_commit.decode('utf-8')
        del proc  # to get rid of the background process
        if repo_commit:
            shaStr = "{}".format(repo_commit.strip())
        else:
            shaStr = 'n/a'
    return shaStr


def _getPlatformString(dist=None):
    """If generic==True then returns empty __build_platform__ string
    """
    if dist == 'bdist':
        # get platform-specific info
        if os.sys.platform == 'darwin':
            OSXver, _, architecture = platform.mac_ver()
            systemInfo = "OSX_%s_%s" % (OSXver, architecture)
        elif os.sys.platform == 'linux':
            import distro
            systemInfo = '%s_%s_%s' % (
                'Linux',
                ':'.join([x for x in distro.linux_distribution() if x != '']),
                platform.release())
        elif os.sys.platform == 'win32':
            ver = os.sys.getwindowsversion()
            if len(ver[4]) > 0:
                systemInfo = "win32_v%i.%i.%i (%s)" % (ver[0], ver[1], ver[2],
                                                       ver[4])
            else:
                systemInfo = "win32_v%i.%i.%i" % (ver[0], ver[1], ver[2])
        else:
            systemInfo = platform.system() + platform.release()
    else:
        systemInfo = "n/a"

    return systemInfo


if __name__ == "__main__":
    createInitFile()



""" Test starting and stopping iohub server
"""
import pytest
from psychopy.tests.utils import skip_under_travis
from psychopy.tests.test_iohub.testutil import startHubProcess, stopHubProcess, getTime

@skip_under_travis
def testDefaultServerLaunch():
    """
    """
    io = startHubProcess()

    # check that a kb and mouse have been created
    keyboard=io.devices.keyboard
    mouse=io.devices.mouse
    exp = io.devices.experiment
    assert keyboard != None
    assert mouse != None
    assert exp != None

    stopHubProcess()




"""ioHub Common Eye Tracker Interface for EyeLink(C) Systems"""
 # Part of the psychopy.iohub library.
 # Copyright (C) 2012-2016 iSolver Software Solutions
 # Distributed under the terms of the GNU General Public License (GPL).


"""
ioHub Common Eye Tracker Interface for EyeLink(C) Systems.
EyeLink(C) calibration graphics implemented using PsychoPy.
"""
# Part of the PsychoPy.iohub library
# Copyright (C) 2012-2016 iSolver Software Solutions
# Distributed under the terms of the GNU General Public License (GPL).
import numpy as np
from PIL import Image, ImageOps
from psychopy import visual
import sys
import tempfile
import os
from ..... import DeviceEvent, Computer
from ......constants import EventConstants, KeyboardConstants
from ......errors import print2err, printExceptionDetailsToStdErr
from ......util import convertCamelToSnake, win32MessagePump
import pylink


class FixationTarget(object):
    def __init__(self, psychopy_eyelink_graphics):
        self.calibrationPointOuter = visual.Circle(
            psychopy_eyelink_graphics.window,
            pos=(0, 0),
            lineWidth=1.0,
            lineColor=psychopy_eyelink_graphics.CALIBRATION_POINT_OUTER_COLOR,
            colorSpace='rgb255',
            fillColor=psychopy_eyelink_graphics.CALIBRATION_POINT_OUTER_COLOR,
            radius=psychopy_eyelink_graphics.CALIBRATION_POINT_OUTER_RADIUS,
            name='CP_OUTER',
            units='pix',
            opacity=1.0,
            interpolate=False)
        self.calibrationPointInner = visual.Circle(
            psychopy_eyelink_graphics.window,
            pos=(0, 0), lineWidth=1.0,
            lineColor=psychopy_eyelink_graphics.CALIBRATION_POINT_INNER_COLOR,
            colorSpace='rgb255',
            fillColor=psychopy_eyelink_graphics.CALIBRATION_POINT_INNER_COLOR,
            radius=psychopy_eyelink_graphics.CALIBRATION_POINT_INNER_RADIUS,
            name='CP_INNER',
            units='pix',
            opacity=1.0,
            interpolate=False)

    def draw(self, pos=None):
        if pos:
            self.calibrationPointOuter.pos = pos
            self.calibrationPointInner.pos = pos
        self.calibrationPointOuter.draw()
        self.calibrationPointInner.draw()


# Intro Screen
class BlankScreen(object):

    def __init__(self, psychopy_win, color):
        self.display_size = psychopy_win.size
        w, h = self.display_size
        self.win = psychopy_win
        self.color = color
        self.background = visual.Rect(self.win, w, h,
                                      lineColor=self.color,
                                      colorSpace='rgb255',
                                      fillColor=self.color,
                                      units='pix',
                                      name='BACKGROUND',
                                      opacity=1.0,
                                      interpolate=False)

    def draw(self):
        self.background.draw()


# Intro Screen
class TextLine(object):

    def __init__(self, psychopy_win):
        self.display_size = psychopy_win.size
        self.win = psychopy_win

        self.textLine = visual.TextStim(
            self.win,
            text='***********************',
            pos=(
                0,
                0),
            height=30,
            color=(
                0,
                0,
                0),
            colorSpace='rgb255',
            opacity=1.0,
            contrast=1.0,
            units='pix',
            ori=0.0,
            antialias=True,
            bold=False,
            italic=False,
            wrapWidth=self.display_size[0] * .8)

    def draw(self, text=None):
        if text:
            self.textLine.text = text
        self.textLine.draw()


# Intro Screen
class IntroScreen(object):
    def __init__(self, psychopy_win):
        self.display_size = psychopy_win.size
        self.window = psychopy_win
        font_height = 24
        space_per_lines = font_height * 2.5
        if self.window.useRetina:
            topline_y = self.window.size[1]/4-font_height*2
        else:
            topline_y = self.window.size[1]/2-font_height*2
        wrap_width = self.window.size[1] * .8
           
        self.introlines = []

        self.introlines.append(visual.TextStim(self.window,
                    text='>>>> Eyelink System Setup:  Keyboard Actions <<<<',
                    pos=(
                        0,
                        topline_y),
                    height=font_height * 1.2,
                    color=(
                        0,
                        0,
                        0),
                    colorSpace='rgb255',
                    opacity=1.0,
                    contrast=1.0,
                    units='pix',
                    ori=0.0,
                    antialias=True,
                    bold=True,
                    italic=False,
                    wrapWidth=wrap_width))

        if self.window.useRetina:
            left_margin = -self.window.size[0]/4
        else:
            left_margin = -self.window.size[0]/2
        left_margin = left_margin *.4
        topline_y = topline_y - space_per_lines/3
        self.introlines.append(visual.TextStim(self.window,
                                               text='* ENTER: Begin Camera Setup Mode',
                                               pos=(left_margin,
                                                    topline_y - space_per_lines * (len(self.introlines))),
                                               height=font_height,
                                               color=(0,
                                                      0,
                                                      0),
                                               colorSpace='rgb255',
                                               opacity=1.0,
                                               contrast=1.0,
                                               units='pix',
                                               ori=0.0,
                                               antialias=True,
                                               bold=False,
                                               italic=False,
                                               alignText='left',
                                               anchorHoriz='left',
                                               wrapWidth=wrap_width))

        self.introlines.append(visual.TextStim(self.window,
                                               text='* C: Start Calibration Procedure',
                                               pos=(left_margin,
                                                    topline_y - space_per_lines * (len(self.introlines))),
                                               height=font_height,
                                               color=(0,
                                                      0,
                                                      0),
                                               colorSpace='rgb255',
                                               opacity=1.0,
                                               contrast=1.0,
                                               units='pix',
                                               ori=0.0,
                                               antialias=True,
                                               bold=True,
                                               italic=False,
                                               alignText='left',
                                               anchorHoriz='left',
                                               wrapWidth=wrap_width))

        self.introlines.append(visual.TextStim(self.window,
                                               text='* V: Start Validation Procedure',
                                               pos=(left_margin,
                                                    topline_y - space_per_lines * (len(self.introlines))),
                                               height=font_height,
                                               color=(0,
                                                      0,
                                                      0),
                                               colorSpace='rgb255',
                                               opacity=1.0,
                                               contrast=1.0,
                                               units='pix',
                                               ori=0.0,
                                               antialias=True,
                                               bold=False,
                                               italic=False,
                                               alignText='left',
                                               anchorHoriz='left',                                               #anchorHoriz='left',
                                               wrapWidth=wrap_width))

        self.introlines.append(visual.TextStim(self.window,
                                               text='* ESCAPE: Exit EyeLink System Setup',
                                               pos=(left_margin,
                                                    topline_y - space_per_lines * (len(self.introlines))),
                                               height=font_height,
                                               color=(0,
                                                      0,
                                                      0),
                                               colorSpace='rgb255',
                                               opacity=1.0,
                                               contrast=1.0,
                                               units='pix',
                                               ori=0.0,
                                               antialias=True,
                                               bold=True,
                                               italic=False,
                                               alignText='left',
                                               anchorHoriz='left',                                               #anchorHoriz='left',
                                               wrapWidth=wrap_width))

        topline_y = topline_y - space_per_lines/3
        self.introlines.append(visual.TextStim(self.window,
                                               text='------ Camera Setup Mode Specific Actions ------',
                                               pos=(0,
                                                    topline_y - space_per_lines * (len(self.introlines))),
                                               height=font_height*1.2,
                                               color=(0,
                                                      0,
                                                      0),
                                               colorSpace='rgb255',
                                               opacity=1.0,
                                               contrast=1.0,
                                               units='pix',
                                               ori=0.0,
                                               antialias=True,
                                               bold=True,
                                               italic=False,
                                               wrapWidth=wrap_width))
        
        topline_y = topline_y - space_per_lines/3
        self.introlines.append(visual.TextStim(self.window,
                                               text='* Left / Right Arrow: Switch Between Camera Views',
                                               pos=(left_margin,
                                                    topline_y - space_per_lines * (len(self.introlines))),
                                               height=font_height,
                                               color=(0,
                                                      0,
                                                      0),
                                               colorSpace='rgb255',
                                               opacity=1.0,
                                               contrast=1.0,
                                               units='pix',
                                               ori=0.0,
                                               antialias=True,
                                               bold=False,
                                               italic=False,
                                               alignText='left',
                                               anchorHoriz='left',                                               #anchorHoriz='left',
                                               wrapWidth=wrap_width))

        self.introlines.append(visual.TextStim(self.window,
                                               text='* A: Auto-Threshold Image',
                                               pos=(left_margin,
                                                    topline_y - space_per_lines * (len(self.introlines))),
                                               height=font_height,
                                               color=(0,
                                                      0,
                                                      0),
                                               colorSpace='rgb255',
                                               opacity=1.0,
                                               contrast=1.0,
                                               units='pix',
                                               ori=0.0,
                                               antialias=True,
                                               bold=False,
                                               italic=False,
                                               alignText='left',
                                               anchorHoriz='left',                                               #anchorHoriz='left',
                                               wrapWidth=wrap_width))

        self.introlines.append(visual.TextStim(self.window,
                                               text='* Up / Down Arrow: Manually Adjust Pupil Threshold',
                                               pos=(left_margin,
                                                    topline_y - space_per_lines * (len(self.introlines))),
                                               height=font_height,
                                               color=(0,
                                                      0,
                                                      0),
                                               colorSpace='rgb255',
                                               opacity=1.0,
                                               contrast=1.0,
                                               units='pix',
                                               ori=0.0,
                                               antialias=True,
                                               bold=False,
                                               italic=False,
                                               alignText='left',
                                               anchorHoriz='left',                                               #anchorHoriz='left',
                                               wrapWidth=wrap_width))

        self.introlines.append(visual.TextStim(self.window,
                                               text='* + or -: Manually Adjust CR Threshold.',
                                               pos=(left_margin,
                                                    topline_y - space_per_lines * (len(self.introlines))),
                                               height=font_height,
                                               color=(0,
                                                      0,
                                                      0),
                                               colorSpace='rgb255',
                                               opacity=1.0,
                                               contrast=1.0,
                                               units='pix',
                                               ori=0.0,
                                               antialias=True,
                                               bold=False,
                                               italic=False,
                                               alignText='left',
                                               anchorHoriz='left',                                               #anchorHoriz='left',
                                               wrapWidth=wrap_width))

    def draw(self):
        for s in self.introlines:
            s.draw()


class EyeLinkCoreGraphicsIOHubPsychopy(pylink.EyeLinkCustomDisplay):
    IOHUB_HEARTBEAT_INTERVAL = 0.050   # seconds between forced run through of
    # micro threads, since one is blocking
    # on camera setup.

    WINDOW_BACKGROUND_COLOR = (128, 128, 128)
    CALIBRATION_POINT_OUTER_RADIUS = 15.0, 15.0
    CALIBRATION_POINT_OUTER_EDGE_COUNT = 64
    CALIBRATION_POINT_OUTER_COLOR = (255, 255, 255)
    CALIBRATION_POINT_INNER_RADIUS = 3.0, 3.0
    CALIBRATION_POINT_INNER_EDGE_COUNT = 32
    CALIBRATION_POINT_INNER_COLOR = (25, 25, 25)

    def __init__(self, eyetrackerInterface, targetForegroundColor=None,
                 targetBackgroundColor=None, screenColor=None,
                 targetOuterDiameter=None, targetInnerDiameter=None):
        pylink.EyeLinkCustomDisplay.__init__(self)

        self._eyetrackerinterface = eyetrackerInterface
        self.tracker = eyetrackerInterface._eyelink
        self._ioKeyboard = None
        self._ioMouse = None

        self.imgstim_size = None
        self.rgb_index_array = None

        self.screenSize = self._eyetrackerinterface._display_device.getPixelResolution()
        self.width = self.screenSize[0]
        self.height = self.screenSize[1]

        self.keys = []
        self.mouse_pos = []
        self.mouse_button_state = 0

        if sys.byteorder == 'little':
            self.byteorder = 1
        else:
            self.byteorder = 0

        EyeLinkCoreGraphicsIOHubPsychopy.CALIBRATION_POINT_OUTER_COLOR = targetForegroundColor
        EyeLinkCoreGraphicsIOHubPsychopy.CALIBRATION_POINT_INNER_COLOR = targetBackgroundColor
        EyeLinkCoreGraphicsIOHubPsychopy.WINDOW_BACKGROUND_COLOR = screenColor
        EyeLinkCoreGraphicsIOHubPsychopy.CALIBRATION_POINT_OUTER_RADIUS = targetOuterDiameter / \
            2.0, targetOuterDiameter / 2.0
        EyeLinkCoreGraphicsIOHubPsychopy.CALIBRATION_POINT_INNER_RADIUS = targetInnerDiameter / \
            2.0, targetInnerDiameter / 2.0

        self.tmp_file = os.path.join(tempfile.gettempdir(), '_eleye.png')

        self.tracker.setOfflineMode()
        self.tracker_version = self.tracker.getTrackerVersion()
        if self.tracker_version >= 3:
            self.tracker.sendCommand('enable_search_limits=YES')
            self.tracker.sendCommand('track_search_limits=YES')
            self.tracker.sendCommand('autothreshold_click=YES')
            self.tracker.sendCommand('autothreshold_repeat=YES')
            self.tracker.sendCommand('enable_camera_position_detect=YES')

        display = self._eyetrackerinterface._display_device
        self.window = visual.Window(display.getPixelResolution(),
                                    monitor=display.getPsychopyMonitorName(),
                                    units=display.getCoordinateType(),
                                    color=self.WINDOW_BACKGROUND_COLOR,
                                    colorSpace='rgb255',
                                    fullscr=True,
                                    allowGUI=False,
                                    screen=display.getIndex()
                                    )

        self.blankdisplay = BlankScreen(
            self.window, self.WINDOW_BACKGROUND_COLOR)
        self.textmsg = TextLine(self.window)
        self.introscreen = IntroScreen(self.window)
        self.fixationpoint = FixationTarget(self)
        self.imagetitlestim = None
        self.eye_image = None
        self.state = None
        self.eye_frame_size = (0, 0)

        self._registerEventMonitors()
        #self._ioMouse.setSystemCursorVisibility(False)
        self._lastMsgPumpTime = Computer.getTime()
        self.clearAllEventBuffers()

    def clearAllEventBuffers(self):
        pylink.flushGetkeyQueue()
        self.tracker.resetData()
        self._iohub_server.eventBuffer.clear()
        for d in self._iohub_server.devices:
            d.clearEvents()

    def _registerEventMonitors(self):
        self._iohub_server = self._eyetrackerinterface._iohub_server

        if self._iohub_server:
            for dev in self._iohub_server.devices:
                if dev.__class__.__name__ == 'Keyboard':
                    kbDevice = dev
                elif dev.__class__.__name__ == 'Mouse':
                    mouseDevice = dev

        if kbDevice:
            eventIDs = []
            for event_class_name in kbDevice.__class__.EVENT_CLASS_NAMES:
                eventIDs.append(
                    getattr(
                        EventConstants,
                        convertCamelToSnake(
                            event_class_name[
                                :-5],
                            False)))

            self._ioKeyboard = kbDevice
            self._ioKeyboard._addEventListener(self, eventIDs)
        else:
            print2err(
                'Warning: elCG could not connect to Keyboard device for events.')

        if mouseDevice:
            eventIDs = []
            for event_class_name in mouseDevice.__class__.EVENT_CLASS_NAMES:
                eventIDs.append(
                    getattr(
                        EventConstants,
                        convertCamelToSnake(
                            event_class_name[
                                :-5],
                            False)))

            self._ioMouse = mouseDevice
            self._ioMouse._addEventListener(self, eventIDs)
        else:
            print2err(
                'Warning: elCG could not connect to Mouse device for events.')

    def _unregisterEventMonitors(self):
        if self._ioKeyboard:
            self._ioKeyboard._removeEventListener(self)
        if self._ioMouse:
            self._ioMouse._removeEventListener(self)

    def _handleEvent(self, event):
        event_type_index = DeviceEvent.EVENT_TYPE_ID_INDEX
        if event[event_type_index] == EventConstants.KEYBOARD_RELEASE:
            from .....keyboard import KeyboardInputEvent
            key_index = KeyboardInputEvent.CLASS_ATTRIBUTE_NAMES.index('key')
            char = event[key_index]
            if isinstance(char, bytes):
                char = str(event[key_index],'utf-8')

            if char:
                char = char.lower()
            else:
                return

            pylink_key = None
            if char == 'escape':
                pylink_key = pylink.ESC_KEY
                self.state = None
            elif char == 'return':
                pylink_key = pylink.ENTER_KEY
                self.state = None
            elif char == ' ':
                pylink_key = ord(char)
            elif char == 'c':
                pylink_key = ord(char)
                self.state = 'calibration'
            elif char == 'v':
                pylink_key = ord(char)
                self.state = 'validation'
            elif char == 'a':
                pylink_key = ord(char)
            elif char == 'pageup':
                pylink_key = pylink.PAGE_UP
            elif char == 'pagedown':
                pylink_key = pylink.PAGE_DOWN
            elif char == '-':
                pylink_key = ord(char)
            elif char == '=':
                pylink_key = ord(char)
            elif char == 'up':
                pylink_key = pylink.CURS_UP
            elif char == 'down':
                pylink_key = pylink.CURS_DOWN
            elif char == 'left':
                pylink_key = pylink.CURS_LEFT
            elif char == 'right':
                pylink_key = pylink.CURS_RIGHT
            else:
                return
            self.keys.append(pylink.KeyInput(pylink_key, 0))

        elif event[event_type_index] == EventConstants.MOUSE_BUTTON_PRESS:
            self.mouse_button_state = 1

        elif event[event_type_index] == EventConstants.MOUSE_BUTTON_RELEASE:
            self.mouse_button_state = 0

        elif event[event_type_index] == EventConstants.MOUSE_MOVE:
            self.mouse_pos = self._ioMouse.getPosition()

    def get_input_key(self):
        if Computer.getTime() - self._lastMsgPumpTime > self.IOHUB_HEARTBEAT_INTERVAL:
            if self._iohub_server:
                win32MessagePump()
                for dm in self._iohub_server.deviceMonitors:
                    dm.device._poll()
                self._iohub_server.processDeviceEvents()
            self._lastMsgPumpTime = Computer.getTime()
        if len(self.keys) > 0:
            k = self.keys
            self.keys = []
            return k
        else:
            return None

    def setup_cal_display(self):
        """Sets up the initial calibration display, which contains a menu with
        instructions."""
        self.blankdisplay.draw()
        self.introscreen.draw()
        self.window.flip()

    def exit_cal_display(self):
        """Exits calibration display."""
        self.clear_cal_display()

    def clear_cal_display(self):
        """Clears the calibration display."""
        self.blankdisplay.draw()
        self.window.flip()

    def erase_cal_target(self):
        """Removes any visible calibration target graphic from display."""
        self.clear_cal_display()

    def draw_cal_target(self, x, y):
        """Draws calibration target."""
        # convert to psychopy pix coords
        x, y = self._eyetrackerinterface._eyeTrackerToDisplayCoords((x,y))

        self.blankdisplay.draw()
        self.fixationpoint.draw((x, y))
        self.window.flip()

    def setup_image_display(self, width, height):
        """Initialize the index array that will contain camera image data."""

        if width and height:
            self.eye_frame_size = (width, height)
            self.clear_cal_display()
            self.last_mouse_state = -1
            if self.rgb_index_array is None:
                self.rgb_index_array = np.zeros((int(height/2), int(width/2)), dtype=np.uint8)

    def exit_image_display(self):
        """Exits the image display."""
        self.clear_cal_display()

    def image_title(self, text):
        """Display the current camera, Pupil, and CR thresholds above the
        camera image when in Camera Setup Mode."""
        if self.imagetitlestim is None:
            self.imagetitlestim = visual.TextStim(
                self.window,
                text=text,
                pos=(
                    0,
                    self.window.size[1] / 2 - 15),
                height=28,
                color=(
                    0,
                    0,
                    0),
                colorSpace='rgb255',
                opacity=1.0,
                contrast=1.0,
                units='pix',
                ori=0.0,
                antialias=True,
                bold=False,
                italic=False,
                anchorVert='top',
                wrapWidth=self.window.size[0] * .8)
        else:
            self.imagetitlestim.setText(text)
        # self.imagetitlestim.draw()

    def draw_image_line(self, width, line, totlines, buff):
        """Collects all lines for an eye image, saves the image, then creates a
        psychopy imagestim from it."""
        for i in range(width):
            try:
                self.rgb_index_array[line - 1, i] = buff[i]
            except Exception as e:
                printExceptionDetailsToStdErr()
                print2err(
                    'FAILED TO DRAW PIXEL TO IMAGE LINE: %d %d' %
                    (line - 1, i))

        # Once all lines have been collected, go through the hoops needed
        # to display the frame as an image; scaled to fit the display
        # resolution.
        if line == totlines:
            try:
                image = Image.fromarray(self.rgb_index_array,
                                           mode='P')
                image.putpalette(self.rgb_pallete)
                image = ImageOps.fit(image, [640, 480])
                if self.eye_image is None:
                    self.eye_image = visual.ImageStim(
                        self.window, image)
                else:
                    self.eye_image.setImage(image)

                # Redraw the Camera Setup Mode graphics
                self.blankdisplay.draw()
                self.eye_image.draw()
                if self.imagetitlestim:
                    self.imagetitlestim.draw()
                self.window.flip()

            except Exception as err:
                print2err('Error during eye image display: ', err)
                printExceptionDetailsToStdErr()

    def set_image_palette(self, r, g, b):
        """Set color palette ued by host pc when sending images.

        Saves the different r,g,b values provided by the eyelink host
        palette. When building up each eye image frame, eyelink sends
        the palette index for each pixel; so an eyelink eye image frame
        can be a 2D lookup array into this palette.

        """
        self.clear_cal_display()
        sz = len(r)
        self.rgb_pallete = np.zeros((sz, 3), dtype=np.uint8)
        i = 0
        while i < sz:
            self.rgb_pallete[i:] = int(r[i]), int(g[i]), int(b[i])
            i += 1

    def alert_printf(self, msg):
        """Prints alert message to psychopy stderr."""
        print2err('eyelink_graphics.alert_printf(): %s' % msg)

###

    def play_beep(self, pylink_sound_index):
        """
        TODO: Plays a sound.
        """
        if pylink_sound_index == pylink.CAL_TARG_BEEP:
            pass
        elif pylink_sound_index == pylink.CAL_ERR_BEEP or pylink_sound_index == pylink.DC_ERR_BEEP:
            self.textmsg.draw('Calibration Failed or Incomplete.')
            self.window.flip()
        elif pylink_sound_index == pylink.CAL_GOOD_BEEP:
            if self.state == 'calibration':
                self.textmsg.draw('Calibration Passed.')
                self.window.flip()
            elif self.state == 'validation':
                self.textmsg.draw('Validation Passed.')
                self.window.flip()
        else:
            pass

    def draw_line(self, x1, y1, x2, y2, color_index):
        """TODO."""
        pass

    def draw_lozenge(self, x, y, width, height, color_index):
        """TODO."""
        pass

    def record_abort_hide(self):
        """TODO."""
        pass

    def get_mouse_state(self):
        """TODO."""
        pass



#!/usr/bin/env python
# -*- coding: utf-8 -*-

# Part of the PsychoPy library
# Copyright (C) 2002-2018 Jonathan Peirce (C) 2019-2021 Open Science Tools Ltd.
# Distributed under the terms of the GNU General Public License (GPL).

"""A Backend class defines the core low-level functions required by a Window
class, such as the ability to create an OpenGL context and flip the window.

Users simply call visual.Window(..., winType='pyglet') and the winType is then
used by backends.getBackend(winType) which will locate the appropriate class
and initialize an instance using the attributes of the Window.
"""

from __future__ import absolute_import, print_function
from builtins import object

import os
import sys
import pygame
from ._base import BaseBackend
import psychopy
from psychopy import core, platform_specific
from psychopy.tools.attributetools import attributeSetter

try:
    import pyglet
    GL = pyglet.gl
except ImportError:
    import OpenGL
    GL = OpenGL


class PygameBackend(BaseBackend):
    """The pygame backend is built on SDL for cross-platform controls
    """
    GL = GL
    winTypeName = 'pygame'

    def __init__(self, win, backendConf=None):
        """Set up the backend window according the params of the PsychoPy win

        Before PsychoPy 1.90.0 this code was executed in Window._setupPygame()

        Parameters
        ----------
        win : `psychopy.visual.Window` instance
            PsychoPy Window (usually not fully created yet).
        backendConf : `dict` or `None`
            Backend configuration options. Options are specified as a dictionary
            where keys are option names and values are settings. This backend
            currently takes no additional settings.

        Examples
        --------
        Create a window using the Pygame backend::

            import psychopy.visual as visual
            win = visual.Window(winType='glfw', backendOptions=options)

        """
        BaseBackend.__init__(self, win)  # sets up self.win=win as weakref

        # pygame.mixer.pre_init(22050,16,2)  # set the values to initialise
        # sound system if it gets used
        pygame.init()
        if win.allowStencil:
            pygame.display.gl_set_attribute(pygame.locals.GL_STENCIL_SIZE, 8)

        try:  # to load an icon for the window
            iconFile = os.path.join(psychopy.__path__[0], 'psychopy.png')
            icon = pygame.image.load(iconFile)
            pygame.display.set_icon(icon)
        except Exception:
            pass  # doesn't matter
        win.useRetina = False
        # these are ints stored in pygame.locals
        winSettings = pygame.OPENGL | pygame.DOUBLEBUF
        if win._isFullScr:
            winSettings = winSettings | pygame.FULLSCREEN
            # check screen size if full screen
            scrInfo = pygame.display.Info()
            win._checkMatchingSizes(win.clientSize,
                                    [scrInfo.current_w,
                                     scrInfo.current_h])
        elif not win.pos:
            # centre video
            os.environ['SDL_VIDEO_CENTERED'] = "1"
        else:
            os.environ['SDL_VIDEO_WINDOW_POS'] = '%i,%i' % (win.pos[0],
                                                            win.pos[1])
        if sys.platform == 'win32':
            os.environ['SDL_VIDEODRIVER'] = 'windib'
        if not win.allowGUI:
            winSettings = winSettings | pygame.NOFRAME
            self.setMouseVisibility(False)
            pygame.display.set_caption('PsychoPy (NB use with allowGUI=False '
                                       'when running properly)')
        else:
            self.setMouseVisibility(True)
            pygame.display.set_caption('PsychoPy')
        self.winHandle = pygame.display.set_mode(win.size.astype('i'),
                                                 winSettings)
        self._frameBufferSize = win.clientSize
        pygame.display.set_gamma(1.0)  # this will be set appropriately later
        # This is causing segfault although it used to be for pyglet only anyway
        # pygame under mac is not syncing to refresh although docs say it should
        # if sys.platform == 'darwin':
        #     platform_specific.syncSwapBuffers(2)

    @property
    def frameBufferSize(self):
        """Framebuffer size (w, h)."""
        return self._frameBufferSize

    def swapBuffers(self, flipThisFrame=True):
        """Do the actual flipping of the buffers (Window will take care of
        additional things like timestamping. Keep this methods as short as poss

        :param flipThisFrame: has no effect on this backend
        """
        if pygame.display.get_init():
            if flipThisFrame:
                pygame.display.flip()
            # keeps us in synch with system event queue
            self.dispatchEvents()
        else:
            core.quit()  # we've unitialised pygame so quit

    def close(self):
        """Close the window and uninitialize the resources
        """
        pygame.display.quit()

    @property
    def shadersSupported(self):
        """This is a read-only property indicating whether or not this backend
        supports OpenGL shaders"""
        return False

    def setMouseVisibility(self, visibility):
        """Set visibility of the mouse to True or False"""
        pygame.mouse.set_visible(visibility)

    # Optional, depending on backend needs

    def dispatchEvents(self):
        """This method is not needed for all backends but for engines with an
        event loop it may be needed to pump for new events (e.g. pyglet)
        """
        pygame.event.pump()

    def onResize(self, width, height):
        """This does nothing; not supported by our pygame backend at the moment
        """
        pass  # the pygame window doesn't currently support resizing

    @attributeSetter
    def gamma(self, gamma):
        self.__dict__['gamma'] = gamma
        # use pygame's own function for this
        pygame.display.set_gamma(gamma[0], gamma[1], gamma[2])

    @attributeSetter
    def gammaRamp(self, gammaRamp):
        """Gets the gamma ramp or sets it to a new value (an Nx3 or Nx1 array)
        """
        self.__dict__['gammaRamp'] = gammaRamp
        # use pygame's own function for this
        pygame.display.set_gamma_ramp(
                gammaRamp[:, 0], gammaRamp[:, 1], gammaRamp[:, 2])

    def setFullScr(self, value):
        """Sets the window to/from full-screen mode"""
        raise NotImplementedError("Toggling fullscreen mode is not currently "
                             "supported on pygame windows")



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
Sound stimuli are currently an area of development in PsychoPy

Previously we used pygame. Now the pyo library is also supported.
On OSX this is an improvement (using coreaudio rather than SDL).
On windows this should help on systems with good sound cards,
but this is yet to be confirmed.

See the demo hardware > testSoundLatency.py
"""

from __future__ import division
from __future__ import print_function

import sys
from psychopy import logging, prefs
logging.console.setLevel(logging.DEBUG)  # get messages about the sound lib as it loads

from psychopy import sound, core

print('Using %s (with %s) for sounds' % (sound.audioLib, sound.audioDriver))

highA = sound.Sound('A', octave=3, sampleRate=44100, secs=0.8, stereo=True)
highA.setVolume(0.8)
tick = sound.Sound(800, secs=0.01, sampleRate=44100, stereo=True)  # sample rate ignored because already set
tock = sound.Sound('600', secs=0.01, sampleRate=44100, stereo=True)

highA.play()
core.wait(0.8)
tick.play()
core.wait(0.4)
tock.play()
core.wait(0.6)

if sys.platform == 'win32':
    ding = sound.Sound('ding')
    ding.play()

    core.wait(1)

    tada = sound.Sound('tada.wav')
    tada.play()

    core.wait(2)
print('done')

core.quit()

# The contents of this file are in the public domain.
